from __future__ import annotations

import logging
import weakref
import pint
import numpy
from silx.gui import qt

from nxtomo.nxobject.nxdetector import ImageKey, FOV
from nxtomo.paths.nxtomo import LATEST_VERSION as _LATEST_NXTOMO_VERSION

from tomwer.core.process.edit.nxtomoeditor import NXtomoEditorKeys
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.gui.utils.buttons import PadlockButton
from tomwer.gui.utils.scandescription import ScanNameLabelAndShape
from tomwer.gui.utils.unitsystem import MetricEntry
from tomwer.gui.edit.nxtomowarmer import NXtomoProxyWarmer

_logger = logging.getLogger(__name__)

_ureg = pint.get_application_registry()


class NXtomoEditorDialog(qt.QDialog):
    """
    Dialog embedding instances of NXtomoEditor and NXtomoProxyWarmer
    """

    def __init__(self, parent=None, hide_lockers=True) -> None:
        super().__init__(parent)
        self.setLayout(qt.QVBoxLayout())

        self.mainWidget = NXtomoEditor(parent=self, hide_lockers=hide_lockers)
        self.layout().addWidget(self.mainWidget)
        self._warmer = NXtomoProxyWarmer(parent=self)
        self.layout().addWidget(self._warmer)

        types = qt.QDialogButtonBox.Ok
        self._buttons = qt.QDialogButtonBox(parent=self)
        self._buttons.setStandardButtons(types)
        self._buttons.button(qt.QDialogButtonBox.Ok).setText("validate")
        self.layout().addWidget(self._buttons)

    # expose API
    def setScan(self, scan: NXtomoScan) -> None:
        self.mainWidget.setScan(scan)
        self._warmer.setScan(scan)

    def hasLockField(self) -> bool:
        return self.mainWidget.hasLockField()

    def getConfiguration(self) -> dict:
        return self.mainWidget.getConfiguration()

    def setConfiguration(self, config: dict) -> None:
        self.mainWidget.setConfiguration(config)

    def getConfigurationForTask(self) -> dict:
        return self.mainWidget.getConfigurationForTask()


class NXtomoEditor(qt.QWidget):
    """
    Widget to edit a set of field from a NXtomo.
    The preliminary goal is to let the user define pixel / voxel position and x and z positions
    in order to simplify stitching down the line

    As energy and field of view was also often requested this part is also editable.

    Each field contains a widget to define it values and a LockButton. The LockButton can be used to automate the
    processing.

    When the scan to edit is set each field widget will be updated **at the condition** the field is not locked.
    Else existing value will be kept.
    """

    sigEditingFinished = qt.Signal()
    """emit when edition is finished"""

    def __init__(self, parent=None, hide_lockers=True):
        super().__init__(parent)
        self._editableWidgets = []
        self._lockerPBs = []
        # list of all lockers
        self._scan = None
        self.setLayout(qt.QVBoxLayout())
        self._scanInfoQLE = ScanNameLabelAndShape(parent=self)
        self.layout().addWidget(self._scanInfoQLE)

        # nxtomo tree
        self._tree = qt.QTreeWidget(self)
        if hide_lockers:
            self._tree.setColumnCount(2)
            self._tree.setHeaderLabels(("entry", "value"))
        else:
            self._tree.setColumnCount(3)
            self._tree.setHeaderLabels(("entry", "value", "lockers"))
            self._tree.header().setStretchLastSection(False)
            self._tree.setColumnWidth(2, 20)
            self._tree.header().setSectionResizeMode(1, qt.QHeaderView.Stretch)
        self.layout().addWidget(self._tree)

        # 1: instrument
        self._instrumentQTWI = qt.QTreeWidgetItem(self._tree)
        self._instrumentQTWI.setText(0, "instrument")
        # handle energy
        self._beamQTWI = qt.QTreeWidgetItem(self._instrumentQTWI)
        self._beamQTWI.setText(0, "beam")
        self._energyQTWI = qt.QTreeWidgetItem(self._beamQTWI)
        self._energyQTWI.setText(0, "energy (keV)")
        self._energyEntry = EnergyEntry("", self)
        self._energyEntry.setPlaceholderText("energy in kev")
        self._tree.setItemWidget(self._energyQTWI, 1, self._energyEntry)
        self._editableWidgets.append(self._energyEntry)
        self._energyLockerLB = PadlockButton(self)
        self._energyLockerLB.setMaximumSize(30, 30)
        self._lockerPBs.append(self._energyLockerLB)
        self._tree.setItemWidget(self._energyQTWI, 2, self._energyLockerLB)

        # 1.1 detector
        self._detectorQTWI = qt.QTreeWidgetItem(self._instrumentQTWI)
        self._detectorQTWI.setText(0, "detector")
        ## pixel size
        self._xDetectorPixelSizeQTWI = qt.QTreeWidgetItem(self._detectorQTWI)
        self._xDetectorPixelSizeQTWI.setText(0, "x pixel size")
        self._xDetectorPixelSizeMetricEntry = MetricEntry("", parent=self)
        self._xDetectorPixelSizeMetricEntry.layout().setContentsMargins(2, 2, 2, 2)
        self._tree.setItemWidget(
            self._xDetectorPixelSizeQTWI, 1, self._xDetectorPixelSizeMetricEntry
        )
        self._editableWidgets.append(self._xDetectorPixelSizeMetricEntry)
        self._xDetectorPixelSizeLB = PadlockButton(self)
        self._xDetectorPixelSizeLB.setMaximumSize(30, 30)
        self._lockerPBs.append(self._xDetectorPixelSizeLB)
        self._tree.setItemWidget(
            self._xDetectorPixelSizeQTWI, 2, self._xDetectorPixelSizeLB
        )

        self._yDetectorPixelSizeQTWI = qt.QTreeWidgetItem(self._detectorQTWI)
        self._yDetectorPixelSizeQTWI.setText(0, "y pixel size")
        self._yDetectorPixelSizeMetricEntry = MetricEntry("", parent=self)
        self._yDetectorPixelSizeMetricEntry.layout().setContentsMargins(2, 2, 2, 2)
        self._tree.setItemWidget(
            self._yDetectorPixelSizeQTWI, 1, self._yDetectorPixelSizeMetricEntry
        )
        self._editableWidgets.append(self._yDetectorPixelSizeMetricEntry)
        self._yDetectorPixelSizeLB = PadlockButton(self)
        self._yDetectorPixelSizeLB.setMaximumSize(30, 30)
        self._lockerPBs.append(self._yDetectorPixelSizeLB)
        self._tree.setItemWidget(
            self._yDetectorPixelSizeQTWI, 2, self._yDetectorPixelSizeLB
        )

        ## sample - distance
        self._sampleDetectorDistanceQTWI = qt.QTreeWidgetItem(self._detectorQTWI)
        self._sampleDetectorDistanceQTWI.setText(0, "sample-detector distance")
        self._sampleDetectorDistanceMetricEntry = MetricEntry("", parent=self)
        self._sampleDetectorDistanceMetricEntry.layout().setContentsMargins(2, 2, 2, 2)
        self._tree.setItemWidget(
            self._sampleDetectorDistanceQTWI, 1, self._sampleDetectorDistanceMetricEntry
        )
        self._editableWidgets.append(self._sampleDetectorDistanceMetricEntry)
        self._sampleDetectorDistanceLB = PadlockButton(self)
        self._sampleDetectorDistanceLB.setMaximumSize(30, 30)
        self._lockerPBs.append(self._sampleDetectorDistanceLB)
        self._tree.setItemWidget(
            self._sampleDetectorDistanceQTWI, 2, self._sampleDetectorDistanceLB
        )

        ## field of view
        self._fieldOfViewQTWI = qt.QTreeWidgetItem(self._detectorQTWI)
        self._fieldOfViewQTWI.setText(0, "field of view")
        self._fieldOfViewCB = qt.QComboBox(self)
        for FOV_item in FOV:
            self._fieldOfViewCB.addItem(FOV_item.value)
        self._tree.setItemWidget(self._fieldOfViewQTWI, 1, self._fieldOfViewCB)
        self._editableWidgets.append(self._fieldOfViewCB)
        self._fieldOfViewLB = PadlockButton(self)
        self._fieldOfViewLB.setMaximumSize(30, 30)
        self._lockerPBs.append(self._fieldOfViewLB)
        self._tree.setItemWidget(self._fieldOfViewQTWI, 2, self._fieldOfViewLB)

        ## x flipped
        self._lrFlippedQTWI = qt.QTreeWidgetItem(self._detectorQTWI)
        self._lrFlippedQTWI.setText(0, "lr flipped")
        self._lrFlippedCB = qt.QCheckBox("", self)
        self._tree.setItemWidget(self._lrFlippedQTWI, 1, self._lrFlippedCB)
        self._editableWidgets.append(self._lrFlippedCB)
        self._lrFlippedLB = PadlockButton(self)
        self._lrFlippedLB.setMaximumSize(30, 30)
        self._lockerPBs.append(self._lrFlippedLB)
        self._tree.setItemWidget(self._lrFlippedQTWI, 2, self._lrFlippedLB)
        ## y flipped
        self._udFlippedQTWI = qt.QTreeWidgetItem(self._detectorQTWI)
        self._udFlippedQTWI.setText(0, "ud flipped")
        self._udFlippedCB = qt.QCheckBox("", self)
        self._tree.setItemWidget(self._udFlippedQTWI, 1, self._udFlippedCB)
        self._editableWidgets.append(self._udFlippedCB)
        self._udFlippedLB = PadlockButton(self)
        self._udFlippedLB.setMaximumSize(30, 30)
        self._lockerPBs.append(self._udFlippedLB)
        self._tree.setItemWidget(self._udFlippedQTWI, 2, self._udFlippedLB)
        # source
        self._sourceQTWI = qt.QTreeWidgetItem(self._instrumentQTWI)
        self._sourceQTWI.setText(0, "source")
        ## source - sample
        self._sampleSourceDistanceQTWI = qt.QTreeWidgetItem(self._sourceQTWI)
        self._sampleSourceDistanceQTWI.setText(0, "sample-source distance")
        self._sampleSourceDistanceMetricEntry = MetricEntry("", parent=self)
        self._sampleSourceDistanceMetricEntry.setToolTip(
            "sample-source distance. Expected to be negative"
        )
        self._sampleSourceDistanceMetricEntry.layout().setContentsMargins(2, 2, 2, 2)
        self._tree.setItemWidget(
            self._sampleSourceDistanceQTWI, 1, self._sampleSourceDistanceMetricEntry
        )
        self._editableWidgets.append(self._sampleSourceDistanceMetricEntry)
        self._sampleSourceDistanceLB = PadlockButton(self)
        self._sampleSourceDistanceLB.setMaximumSize(30, 30)
        self._lockerPBs.append(self._sampleSourceDistanceLB)
        self._tree.setItemWidget(
            self._sampleSourceDistanceQTWI, 2, self._sampleSourceDistanceLB
        )
        # 2: sample
        self._sampleQTWI = qt.QTreeWidgetItem(self._tree)
        self._sampleQTWI.setText(0, "sample")
        # ## pixel size
        self._xSamplePixelSizeQTWI = qt.QTreeWidgetItem(self._sampleQTWI)
        self._xSamplePixelSizeQTWI.setText(0, "x pixel size")
        self._xSamplePixelSizeMetricEntry = MetricEntry("", parent=self)
        self._xSamplePixelSizeMetricEntry.layout().setContentsMargins(2, 2, 2, 2)
        self._tree.setItemWidget(
            self._xSamplePixelSizeQTWI, 1, self._xSamplePixelSizeMetricEntry
        )
        self._editableWidgets.append(self._xSamplePixelSizeMetricEntry)
        self._xSamplePixelSizeLB = PadlockButton(self)
        self._xSamplePixelSizeLB.setMaximumSize(30, 30)
        self._lockerPBs.append(self._xSamplePixelSizeLB)
        self._tree.setItemWidget(
            self._xSamplePixelSizeQTWI, 2, self._xSamplePixelSizeLB
        )

        self._ySamplePixelSizeQTWI = qt.QTreeWidgetItem(self._sampleQTWI)
        self._ySamplePixelSizeQTWI.setText(0, "y pixel size")
        self._ySamplePixelSizeMetricEntry = MetricEntry("", parent=self)
        self._ySamplePixelSizeMetricEntry.layout().setContentsMargins(2, 2, 2, 2)
        self._tree.setItemWidget(
            self._ySamplePixelSizeQTWI, 1, self._ySamplePixelSizeMetricEntry
        )
        self._editableWidgets.append(self._ySamplePixelSizeMetricEntry)
        self._ySamplePixelSizeLB = PadlockButton(self)
        self._ySamplePixelSizeLB.setMaximumSize(30, 30)
        self._lockerPBs.append(self._ySamplePixelSizeLB)
        self._tree.setItemWidget(
            self._ySamplePixelSizeQTWI, 2, self._ySamplePixelSizeLB
        )

        ## propagation distance
        self._propagationDistanceQTWI = qt.QTreeWidgetItem(self._sampleQTWI)
        self._propagationDistanceQTWI.setText(0, "propagation distance")
        self._propagationDistanceMetricEntry = MetricEntry("", parent=self)
        self._propagationDistanceMetricEntry.layout().setContentsMargins(2, 2, 2, 2)
        self._tree.setItemWidget(
            self._propagationDistanceQTWI, 1, self._propagationDistanceMetricEntry
        )
        self._editableWidgets.append(self._propagationDistanceMetricEntry)
        self._propagationDistanceLB = PadlockButton(self)
        self._propagationDistanceLB.setMaximumSize(30, 30)
        self._lockerPBs.append(self._propagationDistanceLB)
        self._tree.setItemWidget(
            self._propagationDistanceQTWI, 2, self._propagationDistanceLB
        )

        ## x translation
        self._xTranslationQTWI = qt.QTreeWidgetItem(self._sampleQTWI)
        self._xTranslationQTWI.setText(0, "x translation")
        self._xTranslationQLE = _TranslationMetricEntry(name="", parent=self)
        self._tree.setItemWidget(self._xTranslationQTWI, 1, self._xTranslationQLE)
        self._editableWidgets.append(self._xTranslationQLE)

        ## z translation
        self._zTranslationQTWI = qt.QTreeWidgetItem(self._sampleQTWI)
        self._zTranslationQTWI.setText(0, "z translation")
        self._zTranslationQLE = _TranslationMetricEntry(name="", parent=self)
        self._tree.setItemWidget(self._zTranslationQTWI, 1, self._zTranslationQLE)
        self._editableWidgets.append(self._zTranslationQLE)

        # set up
        self._instrumentQTWI.setExpanded(True)
        self._sampleQTWI.setExpanded(True)
        self._beamQTWI.setExpanded(True)
        self._detectorQTWI.setExpanded(True)
        self.hideLockers(hide_lockers)

        # connect signal / slot
        self._energyEntry.editingFinished.connect(self._editingFinished)
        self._energyLockerLB.toggled.connect(self._editingFinished)
        self._xDetectorPixelSizeMetricEntry.editingFinished.connect(
            self._editingFinished
        )
        self._xDetectorPixelSizeLB.toggled.connect(self._editingFinished)
        self._yDetectorPixelSizeMetricEntry.editingFinished.connect(
            self._editingFinished
        )
        self._yDetectorPixelSizeLB.toggled.connect(self._editingFinished)
        self._xSamplePixelSizeMetricEntry.editingFinished.connect(self._editingFinished)
        self._xSamplePixelSizeLB.toggled.connect(self._editingFinished)
        self._ySamplePixelSizeMetricEntry.editingFinished.connect(self._editingFinished)
        self._ySamplePixelSizeLB.toggled.connect(self._editingFinished)
        self._sampleDetectorDistanceMetricEntry.editingFinished.connect(
            self._editingFinished
        )
        self._sampleDetectorDistanceLB.toggled.connect(self._editingFinished)
        self._propagationDistanceMetricEntry.editingFinished.connect(
            self._editingFinished
        )
        self._propagationDistanceLB.toggled.connect(self._editingFinished)
        self._fieldOfViewCB.currentIndexChanged.connect(self._editingFinished)
        self._fieldOfViewLB.toggled.connect(self._editingFinished)
        self._lrFlippedCB.toggled.connect(self._editingFinished)
        self._lrFlippedLB.toggled.connect(self._editingFinished)
        self._udFlippedCB.toggled.connect(self._editingFinished)
        self._udFlippedLB.toggled.connect(self._editingFinished)

    def update_tree(self) -> None:
        scan = self.getScan()
        if scan is not None:
            self._updateInstrument()
            self._updateSample()
            self._updateSource()
            self._tree.resizeColumnToContents(0)

            # update items according to NXtomo nexus version
            nexus_version = scan.nexus_version
            if nexus_version is None:
                nexus_version = _LATEST_NXTOMO_VERSION
            # handle nxtomo 1.4 version
            allow_source_sample_distance = bool(nexus_version >= 1.4)
            self._sampleSourceDistanceMetricEntry.setEnabled(
                allow_source_sample_distance
            )
            self._sampleSourceDistanceLB.setEnabled(allow_source_sample_distance)

            # handle nxtomo 1.5 version
            allow_sample_pixel_size_and_prop_distance = bool(nexus_version >= 1.5)
            self._xSamplePixelSizeMetricEntry.setEnabled(
                allow_sample_pixel_size_and_prop_distance
            )
            self._xSamplePixelSizeLB.setEnabled(
                allow_sample_pixel_size_and_prop_distance
            )
            self._ySamplePixelSizeMetricEntry.setEnabled(
                allow_sample_pixel_size_and_prop_distance
            )
            self._ySamplePixelSizeLB.setEnabled(
                allow_sample_pixel_size_and_prop_distance
            )
            self._propagationDistanceMetricEntry.setEnabled(
                allow_sample_pixel_size_and_prop_distance
            )
            self._propagationDistanceLB.setEnabled(
                allow_sample_pixel_size_and_prop_distance
            )

    def _updateInstrument(self) -> None:
        scan = self.getScan()
        if scan is None:
            return
        update_dict = {
            "energy": self._updateEnergy,
            "detector pixel size": self._updateDetectorPixelSize,
            "frame flips": self._updateFlipped,
            "field of view": self._updateFieldOfView,
            "sample-detector distance": self._updateSampleDetectorDistance,
        }
        if scan.nexus_version is None or scan.nexus_version >= 1.4:
            update_dict.update(
                {
                    "sample-source distance": self._updateSampleSourceDistance,
                }
            )
        if scan.nexus_version is None or scan.nexus_version >= 1.5:
            update_dict.update(
                {
                    "propagation distance": self._updatePropagationDistance,
                }
            )
        for name, fct in update_dict.items():
            try:
                fct(scan=scan)
            except Exception as e:
                _logger.error(f"Could not update {name}. Error is {e}")

    def _updateSample(self) -> None:
        scan = self.getScan()
        if scan is None:
            return

        update_dict = {
            "translations": self._updateTranslations,
        }
        if scan.nexus_version is None or scan.nexus_version >= 1.5:
            update_dict.update(
                {
                    "sample pixel size": self._updateSamplePixelSize,
                    "propagation distance": self._updatePropagationDistance,
                }
            )
        for name, fct in update_dict.items():
            try:
                fct(scan=scan)
            except Exception as e:
                _logger.error(f"Fail to update {name}. Error is {e}")

    def _updateSource(self) -> None:
        scan = self.getScan()
        if scan is None:
            return
        if scan.nexus_version is None or scan.nexus_version >= 1.4:
            try:
                self._updateSampleSourceDistance(scan=scan)
            except Exception as e:
                _logger.error(
                    f"Could not update sample - source distance. Error is {e}"
                )

    def _updateTranslations(self, scan: NXtomoScan) -> None:
        assert isinstance(scan, NXtomoScan)

        # note: for now and in order to allow edition we expect to have at most a unique value. Will fail for helicoidal
        def reduce(values):
            if values is None:
                return None
            values = numpy.array(values)
            values = numpy.unique(
                values[scan.image_key_control == ImageKey.PROJECTION.value]
            )
            if values.size == 1:
                return values[0]
            elif values.size == 0:
                return None
            else:
                return f"{values[0]} ... {values[-1]}"

        x_translation = reduce(scan.x_translation)
        z_translation = reduce(scan.z_translation)
        self._xTranslationQLE.setValue(x_translation)
        self._zTranslationQLE.setValue(z_translation)

    def _updateFieldOfView(self, scan: NXtomoScan) -> None:
        if not self._fieldOfViewLB.isLocked():
            # if in ''auto mode: we want to overwrite the NXtomo existing value by the one of the GUI
            idx = self._fieldOfViewCB.findText(FOV.from_value(scan.field_of_view).value)
            if idx > 0:
                self._fieldOfViewCB.setCurrentIndex(idx)

    def _updateFlipped(self, scan: NXtomoScan) -> None:
        flip_lr = scan.detector_is_lr_flip
        flip_ud = scan.detector_is_ud_flip

        if (not self._lrFlippedLB.isLocked()) and flip_lr is not None:
            self._lrFlippedCB.setChecked(flip_lr)
        if (not self._udFlippedLB.isLocked()) and flip_ud is not None:
            self._udFlippedCB.setChecked(flip_ud)

    def _updateSampleDetectorDistance(self, scan: NXtomoScan) -> None:
        if not self._sampleDetectorDistanceLB.isLocked():
            # if in 'auto' mode: we want to overwrite the NXtomo existing value by the one of the GUI
            self._sampleDetectorDistanceMetricEntry.setValue(
                scan.sample_detector_distance
            )

    def _updateSampleSourceDistance(self, scan: NXtomoScan) -> None:
        if not self._sampleSourceDistanceLB.isLocked():
            self._sampleSourceDistanceMetricEntry.setValue(scan.source_sample_distance)

    def _updatePropagationDistance(self, scan: NXtomoScan) -> None:
        if not self._propagationDistanceLB.isLocked():
            # if in 'auto' mode: we want to overwrite the NXtomo existing value by the one of the GUI
            self._propagationDistanceMetricEntry.setValue(scan.propagation_distance)

    def _updateEnergy(self, scan: NXtomoScan) -> None:
        assert isinstance(scan, NXtomoScan)
        if not self._energyLockerLB.isLocked():
            # if in 'auto' mode: we want to overwrite the NXtomo existing value by the one of the GUI
            energy_in_kev: float | None = scan.energy
            if energy_in_kev is not None:
                # move from float to pint.Quantity
                assert not isinstance(energy_in_kev, pint.Quantity)
            self._energyEntry.setValue(energy_in_kev)

    def _updateDetectorPixelSize(self, scan: NXtomoScan) -> None:
        assert isinstance(scan, NXtomoScan)
        if not self._xDetectorPixelSizeLB.isLocked():
            x_pixel_size = scan.detector_x_pixel_size
            self._xDetectorPixelSizeMetricEntry.setValue(x_pixel_size)
        if not self._yDetectorPixelSizeLB.isLocked():
            y_pixel_size = scan.detector_y_pixel_size
            self._yDetectorPixelSizeMetricEntry.setValue(y_pixel_size)

    def _updateSamplePixelSize(self, scan: NXtomoScan) -> None:
        if not self._xSamplePixelSizeLB.isLocked():
            # if in 'auto' mode: we want to overwrite the NXtomo existing value by the one of the GUI
            self._xSamplePixelSizeMetricEntry.setValue(
                scan.get_sample_pixel_size(which="x", fallback_to_det_pixel_size=False)
            )
        if not self._ySamplePixelSizeLB.isLocked():
            # if in 'auto' mode: we want to overwrite the NXtomo existing value by the one of the GUI
            self._ySamplePixelSizeMetricEntry.setValue(
                scan.get_sample_pixel_size(which="y", fallback_to_det_pixel_size=False)
            )

    def _editingFinished(self, *args, **kwargs):
        self.sigEditingFinished.emit()

    def hasLockField(self) -> bool:
        """return True if the widget has at least one lock field"""
        return True in [locker.isLocked() for locker in self._lockerPBs]

    def hideLockers(self, hide: bool) -> None:
        for locker in self._lockerPBs:
            locker.setVisible(not hide)

    def getEditableWidgets(self) -> tuple[qt.QWidget]:
        return tuple(self._editableWidgets)

    def setScan(self, scan: NXtomoScan | None) -> None:
        if scan is None:
            self._scan = scan
        elif not isinstance(scan, NXtomoScan):
            raise TypeError(
                f"{scan} is expected to be an instance of {NXtomoScan}. Not {type(scan)}"
            )
        else:
            self._scan = weakref.ref(scan)
        self._scanInfoQLE.setScan(scan)
        # scan will only be read and not kept
        self.update_tree()

    def getScan(self) -> NXtomoScan | None:
        if self._scan is None or self._scan() is None:
            return None
        else:
            return self._scan()

    def getConfiguration(self) -> dict:
        """
        Return a dict with field full name as key
        and a tuple as value (field_value, is_locked)

        field_value is given in keV for energies else in 'base units'

        limitation: for now sample position are not handled because this is a 'corner case' for now
        """

        def to_base_units_magnitude_if_exists(value: None | pint.Quantity):
            if value is None:
                return None
            else:
                return value.to_base_units().magnitude

        energy = self._energyEntry.getValue()
        if energy is not None:
            assert isinstance(energy, float)

        return {
            NXtomoEditorKeys.ENERGY: (
                energy,
                self._energyLockerLB.isLocked(),
            ),
            NXtomoEditorKeys.DETECTOR_X_PIXEL_SIZE: (
                to_base_units_magnitude_if_exists(
                    self._xDetectorPixelSizeMetricEntry.getValue()
                ),
                self._xDetectorPixelSizeLB.isLocked(),
            ),
            NXtomoEditorKeys.DETECTOR_Y_PIXEL_SIZE: (
                to_base_units_magnitude_if_exists(
                    self._yDetectorPixelSizeMetricEntry.getValue()
                ),
                self._yDetectorPixelSizeLB.isLocked(),
            ),
            NXtomoEditorKeys.SAMPLE_X_PIXEL_SIZE: (
                to_base_units_magnitude_if_exists(
                    self._xSamplePixelSizeMetricEntry.getValue()
                ),
                self._xSamplePixelSizeLB.isLocked(),
            ),
            NXtomoEditorKeys.SAMPLE_Y_PIXEL_SIZE: (
                to_base_units_magnitude_if_exists(
                    self._ySamplePixelSizeMetricEntry.getValue()
                ),
                self._ySamplePixelSizeLB.isLocked(),
            ),
            NXtomoEditorKeys.SAMPLE_DETECTOR_DISTANCE: (
                to_base_units_magnitude_if_exists(
                    self._sampleDetectorDistanceMetricEntry.getValue()
                ),
                self._sampleDetectorDistanceLB.isLocked(),
            ),
            NXtomoEditorKeys.SAMPLE_SOURCE_DISTANCE: (
                to_base_units_magnitude_if_exists(
                    self._sampleSourceDistanceMetricEntry.getValue()
                ),
                self._sampleSourceDistanceLB.isLocked(),
            ),
            NXtomoEditorKeys.FIELD_OF_VIEW: (
                self._fieldOfViewCB.currentText(),
                self._fieldOfViewLB.isChecked(),
            ),
            NXtomoEditorKeys.LR_FLIPPED: (
                self._lrFlippedCB.isChecked(),
                self._lrFlippedLB.isChecked(),
            ),
            NXtomoEditorKeys.UD_FLIPPED: (
                self._udFlippedCB.isChecked(),
                self._udFlippedLB.isChecked(),
            ),
            NXtomoEditorKeys.PROPAGATION_DISTANCE: (
                to_base_units_magnitude_if_exists(
                    self._propagationDistanceMetricEntry.getValue(),
                ),
                self._propagationDistanceLB.isLocked(),
            ),
            NXtomoEditorKeys.X_TRANSLATION: (
                to_base_units_magnitude_if_exists(self._xTranslationQLE.getValue()),
            ),
            NXtomoEditorKeys.Z_TRANSLATION: (
                to_base_units_magnitude_if_exists(self._zTranslationQLE.getValue()),
            ),
        }

    def setConfiguration(self, config: dict) -> None:
        """
        Load given configuration. We expect all quantities to be given:
        * in keV for energies
        * else in base units
        """
        # energy
        energy = config.get(NXtomoEditorKeys.ENERGY, None)
        if energy is not None:
            energy, energy_locked = energy
            assert energy is None or isinstance(
                energy, float
            ), "energy is expected to be dimensionless (float) already in keV"
            self._energyEntry.setValue(energy)
            self._energyLockerLB.setLock(energy_locked)

        # detector pixel size
        detector_x_pixel_size = config.get(NXtomoEditorKeys.DETECTOR_X_PIXEL_SIZE, None)
        if detector_x_pixel_size is not None:
            detector_x_pixel_size, detector_x_pixel_size_locked = detector_x_pixel_size
            self._xDetectorPixelSizeMetricEntry.setValue(
                detector_x_pixel_size, displayed_unit=_ureg.meter
            )
            self._xDetectorPixelSizeLB.setLock(detector_x_pixel_size_locked)

        detector_y_pixel_size = config.get(NXtomoEditorKeys.DETECTOR_Y_PIXEL_SIZE, None)
        if detector_y_pixel_size is not None:
            detector_y_pixel_size, detector_y_pixel_size_locked = detector_y_pixel_size
            self._yDetectorPixelSizeMetricEntry.setValue(
                detector_y_pixel_size, displayed_unit=_ureg.meter
            )
            self._yDetectorPixelSizeLB.setLock(detector_y_pixel_size_locked)

        # sample detector distance
        sample_detector_distance = config.get(
            NXtomoEditorKeys.SAMPLE_DETECTOR_DISTANCE, None
        )
        if sample_detector_distance is not None:
            sample_detector_distance, distance_locked = sample_detector_distance
            self._sampleDetectorDistanceMetricEntry.setValue(
                sample_detector_distance, displayed_unit=_ureg.meter
            )
            self._sampleDetectorDistanceLB.setLock(distance_locked)

        # field of view
        field_of_view = config.get(NXtomoEditorKeys.FIELD_OF_VIEW, None)
        if field_of_view is not None:
            field_of_view, field_of_view_locked = field_of_view
            field_of_view = FOV.from_value(field_of_view)
            self._fieldOfViewCB.setCurrentText(field_of_view.value)
            self._fieldOfViewLB.setLock(field_of_view_locked)

        lr_flipped = config.get(NXtomoEditorKeys.LR_FLIPPED, None)
        if lr_flipped is not None:
            lr_flipped, lr_flipped_locked = lr_flipped
            lr_flipped = lr_flipped in (True, "True", "true")
            self._lrFlippedCB.setChecked(lr_flipped)
            self._lrFlippedLB.setLock(lr_flipped_locked)

        ud_flipped = config.get(NXtomoEditorKeys.UD_FLIPPED, None)
        if ud_flipped is not None:
            ud_flipped, ud_flipped_locked = ud_flipped
            ud_flipped = ud_flipped in (True, "True", "true")
            self._udFlippedCB.setChecked(ud_flipped)
            self._udFlippedLB.setLock(ud_flipped_locked)

        # sample source distance
        sample_source_distance = config.get(
            NXtomoEditorKeys.SAMPLE_SOURCE_DISTANCE, None
        )
        if sample_source_distance is not None:
            sample_source_distance, sample_source_distance_locked = (
                sample_source_distance
            )
            if sample_source_distance >= 0:
                _logger.warning("the sample-source is expected to be negative")
            self._sampleSourceDistanceMetricEntry.setValue(sample_source_distance)
            self._sampleSourceDistanceLB.setLock(sample_source_distance_locked)

        # sample pixel size
        sample_x_pixel_size = config.get(NXtomoEditorKeys.SAMPLE_X_PIXEL_SIZE, None)
        if sample_x_pixel_size is not None:
            sample_x_pixel_size, x_sample_pixel_size_locked = sample_x_pixel_size
            self._xSamplePixelSizeMetricEntry.setValue(
                sample_x_pixel_size, displayed_unit=_ureg.meter
            )
            self._xSamplePixelSizeLB.setLock(x_sample_pixel_size_locked)

        sample_y_pixel_size = config.get(NXtomoEditorKeys.SAMPLE_Y_PIXEL_SIZE, None)
        if sample_y_pixel_size is not None:
            sample_y_pixel_size, sample_y_pixel_size_locked = sample_y_pixel_size
            self._ySamplePixelSizeMetricEntry.setValue(
                sample_y_pixel_size, displayed_unit=_ureg.meter
            )
            self._ySamplePixelSizeLB.setLock(sample_y_pixel_size_locked)

        # propagation distance
        propagation_distance = config.get(NXtomoEditorKeys.PROPAGATION_DISTANCE, None)
        if propagation_distance is not None:
            propagation_distance, propagation_distance_locked = propagation_distance
            self._propagationDistanceMetricEntry.setValue(propagation_distance)
            self._propagationDistanceLB.setChecked(propagation_distance_locked)

    def getConfigurationForTask(self) -> dict:
        """
        default configuration is stored as field: (field_value, filed_is_locked) when the task expects field_key: field_value
        Because we need to be able to reload settings of the LockButton.
        But the task doesn't care about it. She only want to know which field must be edited. So we need to filter the dict value.
        """
        return {key: value[0] for key, value in self.getConfiguration().items()}

    def clear(self) -> None:
        self._tree.clear()


class _TranslationMetricEntry(MetricEntry):
    """
    Widget to define a translation along one axis.

    The behavior is limited at the moment.
    * either the array contains a unique value on the array and a float is displayed
    * either the array contains several unique values and then '...' will be displayed. Users cannot provide an array in this case.
    """

    LOADED_ARRAY = "loaded array"

    class TranslationValidator(qt.QDoubleValidator):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.setNotation(qt.QDoubleValidator.ScientificNotation)

        def validate(self, a0: str, a1: int):
            if "..." in a0:
                return (qt.QDoubleValidator.Acceptable, a0, a1)
            else:
                return super().validate(a0, a1)

    def __init__(self, name, default_unit=_ureg.meter, parent=None):
        super().__init__(name, default_unit=default_unit, parent=parent)
        self._qlePixelSize.setValidator(self.TranslationValidator(self))

    def getValue(self) -> float:
        """

        :return: the value in meter
        """
        if "..." in self._qlePixelSize.text():
            # in this case this is the representation of an array, we don;t wan't to overwrite it
            return self.LOADED_ARRAY
        if self._qlePixelSize.text() in ("unknown", ""):
            return None
        else:
            return float(self._qlePixelSize.text()) * self.getCurrentUnit()


class EnergyEntry(qt.QLineEdit):
    """
    QLineEdit enabling to handle an energy in keV
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setValidator(MetricEntry.DoubleValidator())

    def setValue(self, value_kev: float | None):
        if not isinstance(value_kev, (float, type(None))):
            raise TypeError(
                f"a0 is expected to be {None} or an instance of {float}. Got {type(value_kev)}"
            )

        if value_kev is None:
            value_kev = "unknown"
        else:
            value_kev = str(value_kev)
        super().setText(value_kev)

    def getValue(self) -> float | None:
        """
        Return energy in keV or None
        """
        txt = self.text().replace(" ", "")
        if txt in ("unknown", ""):
            return None
        else:
            return float(txt)



# coding: utf-8
from __future__ import annotations


import logging

from nabu.pipeline.config import get_default_nabu_config
from nabu.pipeline.fullfield.nabu_config import (
    nabu_config as nabu_fullfield_default_config,
)
from silx.gui import icons as silx_icons
from silx.gui import qt
from enum import Enum as _Enum
from nxtomo.nxobject.nxdetector import FOV

from tomwer.core.utils.dictutils import concatenate_dict
from tomwer.core.process.reconstruction.nabu.utils import _NabuMode
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui import icons as tomwer_icons
from tomwer.gui.reconstruction.nabu.nabuflow import NabuFlowArea
from tomwer.gui.reconstruction.nabu.platform import NabuPlatformSettings
from tomwer.gui.utils.buttons import PadlockButton
from tomwer.gui.settings import TAB_LABEL_PLATFORM_SETTINGS

from ...utils.flow import FlowDirection
from tomwer.gui.configuration.action import (
    BasicConfigurationAction,
    ExpertConfigurationAction,
    FilterAction,
    MinimalisticConfigurationAction,
)
from tomwer.gui.configuration.level import ConfigurationLevel
from .nabuconfig import NabuConfiguration

_logger = logging.getLogger(__name__)


class _NabuStages(_Enum):
    INI = "initialization"
    PRE = "pre-processing"
    PHASE = "phase"
    PROC = "processing"
    POST = "post-processing"
    VOLUME = "volume"

    @staticmethod
    def getStagesOrder():
        return (
            _NabuStages.INI,
            _NabuStages.PRE,
            _NabuStages.PHASE,
            _NabuStages.PROC,
            _NabuStages.POST,
        )

    @staticmethod
    def getProcessEnum(stage):
        """Return the process Enum associated to the stage"""
        stage = _NabuStages(stage)
        if stage is _NabuStages.INI:
            raise NotImplementedError()
        elif stage is _NabuStages.PRE:
            return _NabuPreprocessing
        elif stage is _NabuStages.PHASE:
            return _NabuPhase
        elif stage is _NabuStages.PROC:
            return _NabuProcessing
        elif stage is _NabuStages.POST:
            raise NotImplementedError()
        raise NotImplementedError()


class _NabuPreprocessing(_Enum):
    """Define all the preprocessing action possible and the order they
    are applied on"""

    FLAT_FIELD_NORMALIZATION = "flat field normalization"
    CCD_FILTER = "hot spot correction"

    @staticmethod
    def getPreProcessOrder():
        return (
            _NabuPreprocessing.FLAT_FIELD_NORMALIZATION,
            _NabuPreprocessing.CCD_FILTER,
        )


class _NabuPhase(_Enum):
    """Define all the phase action possible and the order they
    are applied on"""

    PHASE = "phase retrieval"
    UNSHARP_MASK = "unsharp mask"
    LOGARITHM = "logarithm"

    @staticmethod
    def getPreProcessOrder():
        return (_NabuPhase.PHASE, _NabuPhase.UNSHARP_MASK, _NabuPhase.LOGARITHM)


class _NabuProcessing(_Enum):
    """Define all the processing action possible"""

    RECONSTRUCTION = 0

    @staticmethod
    def getProcessOrder():
        return _NabuProcessing.RECONSTRUCTION


class _NabuProcess(qt.QWidget):
    def __init__(self, parent):
        qt.QWidget.__init__(self, parent=parent)
        self.setLayout(qt.QGridLayout())
        self._stageCB = qt.QComboBox(parent=self)
        for stage in _NabuStages:
            self._stageCB.addItem(stage.value)

        self.layout().addWidget(qt.QLabel("stage:", self), 0, 0, 1, 1)
        self.layout().addWidget(self._stageCB, 0, 1, 1, 1)

        self._configurationWidget = _NabuConfiguration(parent=parent)
        self.layout().addWidget(self._configurationWidget, 1, 0, 2, 2)


class _NabuConfiguration(qt.QWidget):
    def __init__(self, parent):
        qt.QWidget.__init__(self, parent=parent)
        self.setLayout(qt.QFormLayout)

    def filter(self, stage=None, process=None):
        """Apply a filter on the options to show

        :param stage:
        :param process:
        """
        pass

    def setStages(self, stages: dict) -> None:
        """

        :param stages: contains stages the user can edit and for each
                            stages the associated processes.
        """
        self.clear()
        for stage, processes in stages.items():
            self.addStage(stage=stage, processes=processes)

    def addStage(self, stage, processes):
        stage = _NabuStages(value=stage)
        for process in processes:
            _NabuStages.getProcessEnum(stage=stage)


class NabuDialog(qt.QDialog):
    sigComputationRequested = qt.Signal()
    """Signal emitted when a computation is requested"""

    def __init__(self, parent):
        qt.QDialog.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())

        self._widget = NabuWindow(self)
        self.layout().addWidget(self._widget)

        self._computePB = qt.QPushButton("compute", self)
        self._buttons = qt.QDialogButtonBox(self)
        self._buttons.addButton(self._computePB, qt.QDialogButtonBox.ActionRole)
        self.layout().addWidget(self._buttons)

        # set up

        # expose API
        self.setOutputDir = self._widget.setOutputDir

    def setScan(self, scan):
        self._widget.setScan(scan)

    def getWidget(self):
        return self._widget

    def getConfiguration(self):
        return self._widget.getConfiguration()


class NabuWindow(qt.QMainWindow):
    sigConfigChanged = qt.Signal()
    """Signal emitted when the configuration change"""

    def __init__(self, parent, flow_direction="vertical"):
        qt.QMainWindow.__init__(self, parent=parent)
        self.setWindowFlags(qt.Qt.Widget)

        self._mainWidget = NabuTabWidget(parent=self, flow_direction=flow_direction)
        self.setCentralWidget(self._mainWidget)
        style = qt.QApplication.style()

        # add toolbar
        toolbar = qt.QToolBar(self)
        toolbar.setMovable(False)
        toolbar.setFloatable(False)
        self.addToolBar(qt.Qt.TopToolBarArea, toolbar)

        # add filtering
        self._filterAction = FilterAction(toolbar)
        toolbar.addAction(self._filterAction)
        self._filterAction.triggered.connect(self._filteringChanged)

        # add configuration mode
        self.__configurationModesAction = qt.QAction(self)
        self.__configurationModesAction.setCheckable(False)
        menu = qt.QMenu(self)
        self.__configurationModesAction.setMenu(menu)
        toolbar.addAction(self.__configurationModesAction)

        self.__configurationModesGroup = qt.QActionGroup(self)
        self.__configurationModesGroup.setExclusive(True)
        self.__configurationModesGroup.triggered.connect(self._userModeChanged)

        self._minimalisticAction = MinimalisticConfigurationAction(toolbar)
        menu.addAction(self._minimalisticAction)
        self.__configurationModesGroup.addAction(self._minimalisticAction)
        self._basicConfigAction = BasicConfigurationAction(toolbar)
        menu.addAction(self._basicConfigAction)
        self.__configurationModesGroup.addAction(self._basicConfigAction)
        self._expertConfiguration = ExpertConfigurationAction(toolbar)
        menu.addAction(self._expertConfiguration)
        self.__configurationModesGroup.addAction(self._expertConfiguration)

        # add load option
        self.__loadAction = qt.QAction(self)
        self.__loadAction.setToolTip("load nabu configuration from a text file")
        # load_icon = style.standardIcon(qt.QStyle.SP_)
        load_icon = silx_icons.getQIcon("document-open")
        self.__loadAction.setIcon(load_icon)
        toolbar.addAction(self.__loadAction)
        self.__loadAction.triggered.connect(self._loadParameters)

        # add save option
        self.__saveAction = qt.QAction(self)
        self.__saveAction.setToolTip("save nabu configuration to a text file")
        save_icon = silx_icons.getQIcon("document-save")
        self.__saveAction.setIcon(save_icon)
        toolbar.addAction(self.__saveAction)
        self.__saveAction.triggered.connect(self._saveParameters)

        # reset configuration option
        self.__resetAction = qt.QAction(self)
        self.__resetAction.setToolTip("reset nabu configuration")
        reset_icon = style.standardIcon(qt.QStyle.SP_DialogResetButton)
        self.__resetAction.setIcon(reset_icon)
        toolbar.addAction(self.__resetAction)
        self.__resetAction.triggered.connect(self._resetParameters)

        # toolbar spacer
        self.__tSpacer = qt.QWidget(toolbar)
        self.__tSpacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Preferred)
        toolbar.addWidget(self.__tSpacer)

        # connect signal / slot
        self._mainWidget.sigConfigChanged.connect(self.sigConfigChanged)

        # set up
        self._resetParameters()
        self._filterAction.setChecked(True)
        self._basicConfigAction.setChecked(True)
        self._userModeChanged()

    def getConfiguration(self):
        configuration = self._mainWidget.getConfiguration()
        assert (
            "configuration_level" not in configuration
        ), "configuration_level is a reserved key"
        configuration["configuration_level"] = self.getConfigurationLevel().value
        configuration["mode_locked"] = self.isModeLocked()
        return configuration

    def setConfiguration(self, config):
        conf = config.copy()
        if "configuration_level" in conf:
            configuration_level = conf["configuration_level"]
            del conf["configuration_level"]
        else:
            configuration_level = None
        modeLocked = conf.get("mode_locked", None)
        if modeLocked is not None:
            self.setModeLocked(modeLocked)
        self._mainWidget.setConfiguration(conf)
        if configuration_level is not None:
            self.setConfigurationLevel(level=configuration_level)

    def _saveParameters(self):
        from nabu.pipeline.config import generate_nabu_configfile

        from tomwer.core.process.reconstruction.nabu.nabuslices import (
            interpret_tomwer_configuration,
        )

        # request output file
        config = interpret_tomwer_configuration(
            config=self.getConfiguration(), scan=None
        )
        if len(config) == 0:
            return
        elif len(config) > 1:
            _logger.warning(
                "You requested several paganin values." "Only one will be saved"
            )
        config = config[0][0]
        fname = self.askForNabuconfigFile()
        if fname is None:
            return
        generate_nabu_configfile(
            fname,
            nabu_fullfield_default_config,
            config=config,
            options_level="advanced",
        )

    def _resetParameters(self):
        # reset nabu settings
        default_config = get_default_nabu_config(nabu_fullfield_default_config)
        self._mainWidget.nabuSettingsWidget._configuration._preProcessingWidget._sinoRingsOpts.resetConfiguration()
        default_config["tomwer_slices"] = "middle"
        default_config["preproc"]["ccd_filter_enabled"] = False
        default_config["preproc"]["double_flatfield"] = False
        default_config["preproc"]["flatfield"] = True
        default_config["preproc"]["take_logarithm"] = True
        self.setConfiguration(default_config)

    def _loadParameters(self):
        inputFile = self.askForNabuconfigFile(acceptMode=qt.QFileDialog.AcceptOpen)
        import os

        if inputFile and os.path.exists(inputFile):
            from nabu.pipeline.config import parse_nabu_config_file

            config = parse_nabu_config_file(inputFile)
            self.setConfiguration(config)

    def askForNabuconfigFile(  # pragma: no cover
        self, acceptMode=qt.QFileDialog.AcceptSave
    ):
        dialog = qt.QFileDialog(self)
        dialog.setNameFilters(
            [
                "Configuration files (*.cfg *.conf *.config)",
                "Any files (*)",
            ]
        )

        dialog.setAcceptMode(acceptMode)
        dialog.setFileMode(qt.QFileDialog.AnyFile)

        if not dialog.exec():
            dialog.close()
            return

        filesSelected = dialog.selectedFiles()
        if filesSelected is not None and len(filesSelected) > 0:
            output = filesSelected[0]
            if not output.endswith((".cfg", ".conf", ".config")):
                output = f"{output}.conf"
            return output

    def _userModeChanged(self, *args, **kwargs):
        selectedAction = self.__configurationModesGroup.checkedAction()
        self.__configurationModesAction.setIcon(selectedAction.icon())
        self.__configurationModesAction.setToolTip(selectedAction.tooltip())
        self._mainWidget.setConfigurationLevel(self.getConfigurationLevel())
        self.sigConfigChanged.emit()

    def _filteringChanged(self, *args, **kwargs):
        self._mainWidget.setFilteringActive(self.isFilteringActive())

    def isFilteringActive(self):
        return self._filterAction.isChecked()

    def setConfigurationLevel(self, level):
        level = ConfigurationLevel(level)
        if level == ConfigurationLevel.REQUIRED:
            self._minimalisticAction.setChecked(True)
        elif level == ConfigurationLevel.ADVANCED:
            self._expertConfiguration.setChecked(True)
        elif level == ConfigurationLevel.OPTIONAL:
            self._basicConfigAction.setChecked(True)
        else:
            raise ValueError("Level not recognize")
        self._userModeChanged()

    def getConfigurationLevel(self):
        if self._basicConfigAction.isChecked():
            return ConfigurationLevel.OPTIONAL
        elif self._expertConfiguration.isChecked():
            return ConfigurationLevel.ADVANCED
        elif self._minimalisticAction.isChecked():
            return ConfigurationLevel.REQUIRED
        else:
            raise ValueError("Level not recognize")

    # expose API

    def setModeLocked(self, locked: bool):
        self._mainWidget.nabuSettingsWidget.setModeLocked(locked)

    def isModeLocked(self):
        return self._mainWidget.nabuSettingsWidget.isModeLocked()

    def setOutputDir(self, dir: str):
        self._mainWidget.nabuSettingsWidget.setOutputDir(dir=dir)

    def setScan(self, scan: TomwerScanBase):
        self._mainWidget.nabuSettingsWidget.setScan(scan=scan)

    def getMode(self):
        return self._mainWidget.nabuSettingsWidget.getMode()

    def setMode(self, mode):
        self._mainWidget.nabuSettingsWidget.setMode(mode=mode)

    def hideSlicesInterface(self):
        self._mainWidget.nabuSettingsWidget.hideSlicesInterface()

    def hidePaganinInterface(self):
        self._mainWidget.nabuSettingsWidget.hidePaganinInterface()


class NabuWidget(qt.QWidget):
    """
    Widget containing the entire gui for nabu (control flow + parameters
    settings)
    """

    sigConfigChanged = qt.Signal()
    """Signal emitted when the configuration change"""

    def __init__(self, parent, flow_direction="vertical"):
        qt.QWidget.__init__(self, parent=parent)
        flow_direction = FlowDirection(flow_direction)
        self.setLayout(qt.QGridLayout())
        self._filteringActive = True
        self._configuration_level = ConfigurationLevel.OPTIONAL

        # reconstruction type
        self._widget_recons = qt.QWidget(parent=self)
        self._widget_recons.setLayout(qt.QHBoxLayout())
        self._modeLabel = qt.QLabel("mode:")
        self._widget_recons.layout().addWidget(self._modeLabel)
        self._nabuModeCB = qt.QComboBox(parent=self)
        for mode in _NabuMode:
            self._nabuModeCB.addItem(mode.value)
        self._widget_recons.layout().addWidget(self._nabuModeCB)
        self.layout().addWidget(self._widget_recons, 0, 1, 1, 2)
        # mode lock button
        self._lockModeButton = PadlockButton(parent=self)
        self._lockModeButton.setToolTip(
            "mode can be deduced automatically from the dataset "
            "if it contains information on Field of view. "
            "If you want to enforce the mode you can lock it. "
            "This way automatic deduction will be ignored."
        )
        self._lockModeButton.setMaximumWidth(30)
        self._widget_recons.layout().addWidget(self._lockModeButton)

        # spacer
        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum)
        self.layout().addWidget(spacer, 1, 8, 1, 1)

        # flow
        self._flow = NabuFlowArea(parent=self, direction=flow_direction)
        self._flow.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum)
        self.layout().addWidget(self._flow, 1, 1, 5, 1)

        # nabu configuration
        self._configurationScrollArea = qt.QScrollArea(self)
        self._configurationScrollArea.setWidgetResizable(True)
        self._configuration = NabuConfiguration(parent=self)
        self._configurationScrollArea.setWidget(self._configuration)
        self._configurationScrollArea.setHorizontalScrollBarPolicy(
            qt.Qt.ScrollBarAlwaysOff
        )
        self.layout().addWidget(self._configurationScrollArea, 1, 2, 5, 1)
        self._configuration.setSizePolicy(
            qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum
        )
        self._configurationScrollArea.setSizePolicy(
            qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum
        )

        # expose API
        self.setIniProcessing = self._flow.setIniProcessing
        self.setPreProcessing = self._flow.setPreProcessing
        self.setPhaseProcessing = self._flow.setPhaseProcessing
        self.setProcessing = self._flow.setProcessing
        self.setPostProcessing = self._flow.setPostProcessing

        self.setOutputDir = self._configuration.setOutputDir
        self.getActiveProcess = self._flow.getProcessFocused

        # set up
        pre_processing = ("pre processing",)
        roman_one_icon = tomwer_icons.getQIcon("roman_one")
        self.setPreProcessing(pre_processing, icons=(roman_one_icon,))
        phase_processing = ("phase",)
        roman_two_icon = tomwer_icons.getQIcon("roman_two")
        phase_icons = (roman_two_icon,)
        self.setPhaseProcessing(phase_processing, icons=phase_icons)
        processing = ("reconstruction",)
        processing_icons = (tomwer_icons.getQIcon("roman_three"),)
        self.setProcessing(processes=processing, icons=processing_icons)
        post_processing = ("save",)
        post_processing_icons = (tomwer_icons.getQIcon("roman_four"),)
        self.setPostProcessing(post_processing, icons=post_processing_icons)
        index_mode = self._nabuModeCB.findText(_NabuMode.FULL_FIELD.value)
        assert index_mode >= 0, "full filed should be registered in the widget"
        self._nabuModeCB.setCurrentIndex(index_mode)
        self._flow.setMaximumWidth(240)

        # signal / slot connections
        self._flow.sigConfigurationChanged.connect(self._processSelectionChanged)
        self._flow.sigResetConfiguration.connect(self._processSelectionChanged)
        self._configuration.sigConfChanged.connect(self._triggerConfigChanged)
        self._nabuModeCB.currentIndexChanged.connect(self._triggerConfigChanged)

    def _triggerConfigChanged(self, *args, **kwargs):
        self.sigConfigChanged.emit()

    def isModeLocked(self) -> True:
        """
        Return True if the mode is locked (in this case there is no automatic detection of it)
        """
        return self._lockModeButton.isLocked()

    def setModeLocked(self, locked: bool):
        if not isinstance(locked, bool):
            raise TypeError(f"locked is expeced to be a bool not {type(locked)}")
        self._lockModeButton.setLock(locked)

    def setScan(self, scan):
        """Tune the configuration from scan information if possible"""
        assert isinstance(scan, TomwerScanBase)
        if scan.field_of_view is not None and not self.isModeLocked():
            if scan.field_of_view == FOV.FULL:
                self.setMode(_NabuMode.FULL_FIELD)
            elif scan.field_of_view == FOV.HALF:
                self.setMode(_NabuMode.HALF_ACQ)
        if scan.sa_delta_beta_params is not None:
            db = scan.sa_delta_beta_params.selected_delta_beta_value
            if db is not None:
                self.setDeltaBetaValue(db)

    def getConfiguration(self):
        conf = self._configuration.getConfiguration()
        enable_ht = int(self.getMode() is _NabuMode.HALF_ACQ)
        conf["reconstruction"]["enable_halftomo"] = enable_ht
        return conf

    def setConfiguration(self, config):
        if "reconstruction" in config and "enable_halftomo" in config["reconstruction"]:
            if config["reconstruction"]["enable_halftomo"] == 1:
                index = self._nabuModeCB.findText(_NabuMode.HALF_ACQ.value)
            else:
                index = self._nabuModeCB.findText(_NabuMode.FULL_FIELD.value)
            self._nabuModeCB.setCurrentIndex(index)
        self._configuration.setConfiguration(config=config)

    def setDeltaBetaValue(self, value):
        self._configuration.setDeltaBetaValue(value=value)

    def hideSlicesInterface(self):
        self._configuration.hideSlicesInterface()

    def hidePaganinInterface(self):
        self._configuration.hidePaganinInterface()

    def getMode(self):
        return _NabuMode(self._nabuModeCB.currentText())

    def setMode(self, mode):
        mode = _NabuMode(mode)
        idx = self._nabuModeCB.findText(mode.value)
        self._nabuModeCB.setCurrentIndex(idx)

    def _processSelectionChanged(self, *arg):
        if self.isConfigFiltered():
            self.updateConfigurationFilter()

    def isConfigFiltered(self):
        return self._filteringActive

    def updateConfigurationFilter(self, *args, **kwargs):
        if self.isConfigFiltered():
            stage, option = self.getActiveProcess()
        else:
            stage = None
            option = None
        self._configuration.applyFilter(stage=stage, option=option)
        self._configuration.setConfigurationLevel(self.getConfigurationLevel())
        # force scroll bar to update
        self._configurationScrollArea.updateGeometry()

    def setConfigurationLevel(self, level):
        level = ConfigurationLevel(level)
        self._configuration_level = level
        self.updateConfigurationFilter()

    def getConfigurationLevel(self):
        return self._configuration_level

    def setFilteringActive(self, active):
        self._filteringActive = active
        self.updateConfigurationFilter()


class NabuTabWidget(qt.QTabWidget):
    """
    Widget that group Nabu reconstruction settings and (local) platform settings
    """

    sigConfigChanged = qt.Signal()

    def __init__(self, parent=None, flow_direction: str = "vertical"):
        super().__init__(parent=parent)
        self._nabuSettingsWidget = NabuWidget(
            parent=self, flow_direction=flow_direction
        )
        self.addTab(self._nabuSettingsWidget, "reconstruction settings")
        self._platformSettingsWidget = NabuPlatformSettings(parent=self)
        self.addTab(self._platformSettingsWidget, TAB_LABEL_PLATFORM_SETTINGS)

        # connect signal / slot
        self._nabuSettingsWidget.sigConfigChanged.connect(self.sigConfigChanged)
        self._platformSettingsWidget.sigConfigChanged.connect(self.sigConfigChanged)

    def getConfiguration(self) -> dict:
        config = self._nabuSettingsWidget.getConfiguration()
        config = concatenate_dict(
            config, self._platformSettingsWidget.getConfiguration()
        )
        return config

    def setConfiguration(self, config: dict):
        self._nabuSettingsWidget.setConfiguration(config=config)
        self._platformSettingsWidget.setConfiguration(config=config)

    def setConfigurationLevel(self, level):
        self._nabuSettingsWidget.setConfigurationLevel(level=level)
        self._platformSettingsWidget.setConfigurationLevel(level=level)

    def getConfigurationLevel(self):
        return self._nabuSettingsWidget.getConfigurationLevel()

    @property
    def nabuSettingsWidget(self):
        return self._nabuSettingsWidget

    @property
    def platformSettings(self):
        return self._platformSettingsWidget



"""
(Ewoks)Orange Widgets for NXtomo edition
"""

NAME = "edit HDF5 - NXTomo"

ID = "orangecontrib.tomwer.widgets.edit"

DESCRIPTION = "widgets for editing NXTomo"

LONG_DESCRIPTION = "Contains widgets for editing NXtomo entries"

ICON = "../../widgets/icons/tomwer_edit.png"

BACKGROUND = "#C0CCFF"

PRIORITY = 8



# coding: utf-8
"""
Applications and library for tomography processing.
Provides ewoks tasks, several GUI for nabu as several utils.
"""

from __future__ import absolute_import, division, print_function

from tomwer.core.log import processlog  # noqa F401
from tomwer.version import version as __version  # noqa F401

try:
    # try to first load hdf5plugin before h5py
    import hdf5plugin  # noqa F401
except ImportError:
    pass


__version__ = __version



"""Define some processing stack"""

from __future__ import annotations

from collections import deque
from tomwer.core.tomwer_object import TomwerObject
from silx.gui.utils import blockSignals

from processview.core.manager import DatasetState, ProcessManager
from processview.core.superviseprocess import SuperviseProcess
from silx.gui import qt


class ProcessingThread(qt.QThread):
    """Class for running some processing"""

    sigComputationStarted = qt.Signal()
    """Signal emitted when a computation is started"""

    sigError = qt.Signal(str)
    """Error found during processing. Parameter is the error raised"""


class FIFO(SuperviseProcess):
    """Processing Queue with a First In, First Out behavior"""

    sigComputationStarted = qt.Signal(object)
    """Signal emitted when a computation is started"""
    sigComputationEnded = qt.Signal(object, object)
    """Signal emitted when a computation is ended. First parameter is mandatory and
    is the scan computed. Second one is an optional FutureTomwerScan"""

    def __init__(self, process_id=None):
        SuperviseProcess.__init__(self, process_id=process_id)
        self._errors: list[qt.QErrorMessage] = []
        self._deque = deque()
        self._computationThread = self._create_processing_thread(process_id=process_id)
        assert isinstance(self._computationThread, ProcessingThread)

        """scan process by the thread"""
        self._data_currently_computed = None
        """Scan computed currently"""
        self._processing = False

        # connect signal / slot
        self._computationThread.sigComputationStarted.connect(
            self._start_threaded_computation
        )
        self._computationThread.sigError.connect(self._displayError)

    @property
    def data_currently_computed(self) -> TomwerObject | None:
        return self._data_currently_computed

    def add(self, data, configuration=None, callback=None):
        """
        add a scan to process

        :param data: data to process
        :param configuration: configuration of the process
        :param callback: function to call once the processing is Done
        """
        try:
            if self.process_id not in (None, -1):
                ProcessManager().notify_dataset_state(
                    dataset=data,
                    process=self,
                    state=DatasetState.PENDING,
                )
        except Exception:
            pass

        self.append((data, configuration, callback))
        if self.can_process_next():
            self._process_next()

    def _process(self, data, configuration, callback):
        raise NotImplementedError("Virtual class")

    def _process_next(self):
        if len(self) == 0:
            return

        self._processing = True
        data, configuration, callback = self.pop()
        self._process(data=data, configuration=configuration or {}, callback=callback)

    def can_process_next(self):
        """
        :return: True if the computation thread is ready to compute a new axis position
        """
        return not self._processing

    def _end_computation(self, data, future_tomo_obj, callback):
        """
        callback when the computation thread is finished

        :param scan: pass if no call to '_computationThread is made'
        """
        if callback is not None:
            callback()
        sender = self.sender()  # pylint: disable=E1101
        if hasattr(sender, "future_tomo_obj"):
            future_tomo_obj = sender.future_tomo_obj
        else:
            future_tomo_obj = None

        self.sigComputationEnded.emit(data, future_tomo_obj)
        self._processing = False
        if self.can_process_next():
            self._process_next()

    def _create_processing_thread(self, process_id=None) -> ProcessingThread:
        raise NotImplementedError("Virtual class")

    def is_computing(self):
        """Return True if processing thread is running (mean that computation
        is on going)"""
        return self._processing

    def wait_computation_finished(self):
        """
        Wait until the computation is finished
        """
        if self._processing:
            self._computationThread.wait()

    def _start_threaded_computation(self, *args, **kwargs):
        self.sigComputationStarted.emit(self._data_currently_computed)

    def _end_threaded_computation(self, callback=None):
        sender = self.sender()  # pylint: disable=E1101
        if hasattr(sender, "future_tomo_obj"):
            future_tomo_obj = sender.future_tomo_obj
        else:
            future_tomo_obj = None
        self._end_computation(
            data=self._data_currently_computed,
            callback=callback,
            future_tomo_obj=future_tomo_obj,
        )

    def cancel(self):
        if self._computationThread.isRunning():
            with blockSignals(self._computationThread):
                self._computationThread.cancel()
                # stop stack
                self.stop()
                # emit next
                self.sigComputationEnded.emit(None, None)

    def stop(self):
        self._computationThread.wait()
        self._processing = False
        self._data_currently_computed = None

    def _displayError(self, error_msg: str):
        err = qt.QMessageBox()
        err.setText(error_msg)
        err.setIcon(qt.QMessageBox.Warning)
        self._errors.append(err)
        err.setModal(False)
        err.show()

    # expose deque API
    def append(self, value):
        self._deque.append(value)

    def clear(self):
        self._deque.clear()

    def pop(self):
        return self._deque.pop()

    def remove(self, value):
        return self._deque.remove(value)

    def __len__(self):
        return len(self._deque)

    def __contains__(self, value):
        return value.get_identifier().to_str() in [
            scan.get_identifier().to_str() for scan in self._deque
        ]



from __future__ import annotations

import weakref
from silx.gui import qt

from tomwer.gui import icons


class ShowCoordinateSystemAction(qt.QAction):
    """
    QAction to show / hide the coordinate system vignette of the plots
    """

    STATE = None
    """Lazy loaded states used to feed the coordinate system"""

    def __init__(self, plots: tuple, parent=None):
        icon = icons.getQIcon("show_3D_coordinate_system")
        text = "Toggle 3D coordinate system"
        super().__init__(parent=parent, icon=icon, text=text)

        self.setToolTip("Toggle 3D coordinate system of slice plots")
        self.setCheckable(True)
        self.setChecked(True)

        self._plots = [weakref.ref(plot) for plot in plots]

        self.triggered[bool].connect(self._updateCoordinateSystemVisibility)

    def _updateCoordinateSystemVisibility(self, visible):
        """Handle Plot set keep aspect ratio signal"""
        for ref_plot in self._plots:
            plot = ref_plot()
            if plot:
                plot.setCoordinateSystemVisible(visible)



from __future__ import annotations

import enum
import logging

from silx.gui import qt

from tomwer.core.process.reconstruction.axis import mode as axis_mode
from tomwer.gui.utils.qt_utils import block_signals
from tomwer.gui.utils.step import StepSizeSelectorWidget
from tomwer.synctools.axis import QAxisRP
from tomwer.gui.settings import EDITING_BACKGROUND_COLOR

from .AxisOptionsWidget import AxisOptionsWidget
from .CalculationWidget import CalculationWidget
from .InputWidget import InputWidget


_logger = logging.getLogger(__name__)


@enum.unique
class ShiftMode(enum.Enum):
    x_only = 0
    y_only = 1
    x_and_y = 2


class AxisSettingsWidget(qt.QWidget):
    """
    Widget to define settings for COR search (which algorithm to use...)
    """

    sigShiftChanged = qt.Signal(float, float)
    """Signal emitted when requested shift changed. Parameter is x, y"""

    sigModeLockChanged = qt.Signal(bool)
    """Signal emitted when the mode is lock or unlock"""

    sigResetZoomRequested = qt.Signal()
    """Signal emitted when request a zoom reset from the plot"""

    sigSubSamplingChanged = qt.Signal()
    """Signal emitted when sub-sampling change"""

    sigUrlChanged = qt.Signal()
    """Signal emit when frames urls changed"""

    sigModeChanged = qt.Signal(str)
    """Signal emit when mode (algorithm) is changed"""

    sigRoiChanged = qt.Signal(object)
    """Signal emitted when the ROI changed"""

    sigAuto = qt.Signal()
    """Signal emitted when the auto button is activated"""

    def __init__(self, parent, reconsParams):
        assert isinstance(reconsParams, QAxisRP)
        qt.QWidget.__init__(self, parent)
        self._xShift = 0
        self._yShift = 0
        self._recons_params = reconsParams
        self._axisParams = None

        self.setLayout(qt.QVBoxLayout())

        self._manualSelectionWidget = ManualAxisSelectionWidget(
            parent=self, shift_mode=ShiftMode.x_only
        )
        self._manualSelectionWidget.layout().setContentsMargins(0, 0, 0, 0)

        self._displacementSelector = self._manualSelectionWidget._displacementSelector
        self._shiftControl = self._manualSelectionWidget._shiftControl
        self._roiControl = self._manualSelectionWidget._roiControl
        self._imgOpts = self._manualSelectionWidget._imgOpts

        self._mainWidget = AxisSettingsTabWidget(
            parent=self,
            recons_params=self._recons_params,
        )
        # append the Manual selection to the AxisTabWidget to have all widget related to axis settings together
        self._mainWidget._calculationWidget.layout().addWidget(
            self._manualSelectionWidget
        )

        self.layout().addWidget(self._mainWidget)

        # signal / slot connection
        self._shiftControl.sigShiftLeft.connect(self._incrementLeftShift)
        self._shiftControl.sigShiftRight.connect(self._incrementRightShift)
        self._shiftControl.sigShiftTop.connect(self._incrementTopShift)
        self._shiftControl.sigShiftBottom.connect(self._incrementBottomShift)
        self._shiftControl.sigReset.connect(self._resetShift)
        self._shiftControl.sigShiftChanged.connect(self._setShiftAndSignal)
        self._mainWidget.sigLockModeChanged.connect(self._modeLockChanged)
        self._manualSelectionWidget.sigResetZoomRequested.connect(
            self._requestZoomReset
        )
        self._imgOpts.sigSubSamplingChanged.connect(self.sigSubSamplingChanged)
        self._mainWidget.sigUrlChanged.connect(self.sigUrlChanged)
        self._roiControl.sigRoiChanged.connect(self.sigRoiChanged)
        self._shiftControl.sigAuto.connect(self.sigAuto)
        self._mainWidget.sigModeChanged.connect(self.sigModeChanged)

        # set up interface
        self.setAxisParams(self._recons_params)

    def setScan(self, scan):
        self._mainWidget.setScan(scan=scan)
        self._roiControl.setScan(scan=scan)

    def manual_uses_full_image(self, value):
        self._roiControl.manual_uses_full_image(value)

    def _incrementLeftShift(self):
        self._incrementShift("left")

    def _incrementRightShift(self):
        self._incrementShift("right")

    def _incrementTopShift(self):
        self._incrementShift("top")

    def _incrementBottomShift(self):
        self._incrementShift("bottom")

    def _setShiftAndSignal(self, x, y):
        if x == self._xShift and y == self._yShift:
            return
        self.setShift(x, y)
        self._shiftControl._updateShiftInfo(x=x, y=y)
        self.sigShiftChanged.emit(x, y)

    def setAxisParams(self, axis):
        if axis == self._axisParams:
            return
        assert isinstance(axis, QAxisRP)
        with block_signals(self):
            if self._axisParams:
                self._axisParams.sigChanged.disconnect(self._updateAxisView)
            self._axisParams = axis
            self.setXShift(self._axisParams.relative_cor_value)
            self._mainWidget.setAxisParams(self._axisParams)
            self._updateAxisView()
            self._axisParams.sigChanged.connect(self._updateAxisView)

    def _modeLockChanged(self, lock):
        self.sigModeLockChanged.emit(lock)

    def setMode(self, mode):
        with block_signals(self._axisParams):
            self._axisParams.mode = mode
            self._mainWidget._calculationWidget.setMode(mode)
            self._mainWidget._optionsWidget.setMode(mode)
            self._updateAxisView()
        self._axisParams.sigChanged.emit()

    def _updateAxisView(self):
        with block_signals(self._axisParams):
            if self._axisParams.relative_cor_value not in (None, "..."):
                self.setXShift(self._axisParams.relative_cor_value)

        self._manualSelectionWidget.setVisible(
            self._axisParams.mode is axis_mode.AxisMode.manual
        )

    def getAxisParams(self):
        return self._axisParams

    def _incrementShift(self, direction):
        assert direction in ("left", "right", "top", "bottom")
        if direction == "left":
            self.setXShift(self._xShift - self.getShiftStep())
        elif direction == "right":
            self.setXShift(self._xShift + self.getShiftStep())
        elif direction == "top":
            self.setYShift(self._yShift + self.getShiftStep())
        else:
            self.setYShift(self._yShift - self.getShiftStep())

        self._shiftControl._updateShiftInfo(x=self._xShift, y=self._yShift)

    def _resetShift(self):
        with block_signals(self._axisParams):
            self.setXShift(0)
            self.setYShift(0)
            self._shiftControl._updateShiftInfo(x=self._xShift, y=self._yShift)

        self.sigShiftChanged.emit(self._xShift, self._yShift)

    def getXShift(self):
        if self._xShift == "...":
            return 0
        return self._xShift

    def getYShift(self):
        if self._yShift == "...":
            return 0
        return self._yShift

    def setXShift(self, x: float):
        self.setShift(x=x, y=self._yShift)

    def setYShift(self, y):
        self.setShift(x=self._xShift, y=y)

    def setShift(self, x, y):
        if x == self._xShift and y == self._yShift:
            return
        self._xShift = x if x is not None else 0.0
        self._yShift = y if y is not None else 0.0
        if self._axisParams:
            with block_signals(self._axisParams):
                self._axisParams.set_relative_value(x)
        self._shiftControl._updateShiftInfo(x=self._xShift, y=self._yShift)
        if not isinstance(self._xShift, str):
            # filter `...` and `?` values (used for issues or processing)
            self.sigShiftChanged.emit(self._xShift, self._yShift)

    def reset(self):
        with block_signals(self):
            self.setShift(0, 0)
        self.sigShiftChanged.emit(self._xShift, self._yShift)

    def setLocked(self, locked):
        self._mainWidget.setEnabled(not locked)

    def _requestZoomReset(self):
        self.sigResetZoomRequested.emit()

    # expose API

    def setEstimatedCor(self, value):
        self._mainWidget.setEstimatedCorValue(value=value)

    def updateXRotationAxisPixelPositionOnNewScan(self) -> bool:
        return self._mainWidget.updateXRotationAxisPixelPositionOnNewScan()

    def setUpdateXRotationAxisPixelPositionOnNewScan(self, update: bool):
        self._mainWidget.setUpdateXRotationAxisPixelPositionOnNewScan(update=update)

    def getEstimatedCor(self):
        return self._mainWidget.getEstimatedCor()

    def getROIDims(self):
        return self._roiControl.getROIDims()

    def getShiftStep(self):
        return self._displacementSelector.getStepSize()

    def setShiftStep(self, value):
        self._displacementSelector.setStepSize(value=value)

    def getROIOrigin(self):
        return self._roiControl.getROIOrigin()

    def getImgSubSampling(self):
        return self._imgOpts.getSubSampling()

    def getMode(self):
        return self._mainWidget.getMode()

    def isModeLock(self):
        return self._mainWidget.isModeLock()

    def setModeLock(self, mode):
        return self._mainWidget.setModeLock(mode=mode)

    def isYAxisInverted(self) -> bool:
        return self._mainWidget.isYAxisInverted()

    def setYAxisInverted(self, checked: bool):
        return self._mainWidget.setYAxisInverted(checked=checked)


class ManualAxisSelectionWidget(qt.QWidget):
    sigResetZoomRequested = qt.Signal()
    """Signal emitted when a zoom request is necessary (when change to full
    image)"""

    def __init__(self, parent, shift_mode):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())
        self._displacementSelector = StepSizeSelectorWidget(
            parent=self,
            fine_value=0.1,
            medium_value=1.0,
            rough_value=None,
            dtype=float,
        )
        self.layout().addWidget(self._displacementSelector)

        self._shiftControl = _ShiftControl(parent=self, shift_mode=shift_mode)
        self.layout().addWidget(self._shiftControl)

        self._roiControl = _ROIControl(parent=self)
        self.layout().addWidget(self._roiControl)

        self._imgOpts = _ImgOpts(parent=self)
        self.layout().addWidget(self._imgOpts)

        # connect signal / slot
        self._roiControl.sigResetZoomRequested.connect(self.sigResetZoomRequested)


class _ROIControl(qt.QGroupBox):
    """
    Widget used to define the ROI on images to compare
    """

    sigRoiChanged = qt.Signal(object)
    """Signal emitted when the ROI changed"""
    sigResetZoomRequested = qt.Signal()
    """Signal emitted when a zoom request is necessary (when change to full
    image)"""

    def __init__(self, parent):
        qt.QGroupBox.__init__(self, "ROI selection", parent)
        self.setLayout(qt.QVBoxLayout())

        self._buttonGrp = qt.QButtonGroup(parent=self)
        self._buttonGrp.setExclusive(True)

        self._roiWidget = qt.QWidget(parent=self)
        self._roiWidget.setLayout(qt.QHBoxLayout())
        self._roiWidget.layout().setContentsMargins(0, 0, 0, 0)
        self._fullImgButton = qt.QRadioButton("full image", parent=self)
        self._buttonGrp.addButton(self._fullImgButton)
        self.layout().addWidget(self._fullImgButton)
        self._roiButton = qt.QRadioButton("ROI", parent=self._roiWidget)
        self._roiWidget.layout().addWidget(self._roiButton)
        self._buttonGrp.addButton(self._roiButton)
        self._roiDefinition = _ROIDefinition(parent=self)
        self._roiWidget.layout().addWidget(self._roiDefinition)
        self.layout().addWidget(self._roiWidget)

        # connect signal / Slot
        self._roiButton.toggled.connect(self._roiDefinition.setEnabled)
        self._fullImgButton.toggled.connect(self.sigResetZoomRequested)
        self._roiButton.toggled.connect(self.sigResetZoomRequested)
        self._roiDefinition.sigRoiChanged.connect(self.sigRoiChanged)

        # setup for full image
        self._fullImgButton.setChecked(True)

    def getROIDims(self):
        if self._roiButton.isChecked():
            return self._roiDefinition.getROIDims()
        else:
            return None

    def manual_uses_full_image(self, activate):
        if activate:
            self._fullImgButton.setChecked(True)
        else:
            self._roiButton.setChecked(True)

    # expose API
    def getROIOrigin(self):
        return self._roiDefinition.getROIOrigin()

    def setLimits(self, width, height):
        self._roiDefinition.setLimits(width=width, height=height)

    def setScan(self, scan):
        self._roiDefinition.setScan(scan=scan)


class _ROIDefinition(qt.QWidget):
    """
    Widget used to define ROI width and height.

    :note: emit ROI == None if setDisabled
    """

    sigRoiChanged = qt.Signal(object)
    """Signal emitted when the ROI changed"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QGridLayout())
        self._already_set = False

        # width & height
        self.layout().addWidget(qt.QLabel("dims", self), 0, 0)
        self._widthSB = qt.QSpinBox(parent=self)
        self._widthSB.setSingleStep(2)
        self._widthSB.setMaximum(10000)
        self._widthSB.setSuffix(" px")
        self._widthSB.setPrefix("w: ")
        self._widthSB.setToolTip("ROI width")
        self.layout().addWidget(self._widthSB, 0, 1)
        self._heightSB = qt.QSpinBox(parent=self)
        self._heightSB.setSingleStep(2)
        self._heightSB.setSuffix(" px")
        self._heightSB.setPrefix("h: ")
        self._heightSB.setToolTip("ROI height")
        self._heightSB.setMaximum(10000)
        self.layout().addWidget(self._heightSB, 0, 2)

        # origin x and y position
        self.layout().addWidget(qt.QLabel("origin", self), 1, 0)
        self._xOriginSB = qt.QSpinBox(parent=self)
        self._xOriginSB.setSingleStep(10)
        self._xOriginSB.setMaximum(10000)
        self._xOriginSB.setPrefix("x: ")
        self.layout().addWidget(self._xOriginSB, 1, 1)
        self._yOriginSB = qt.QSpinBox(parent=self)
        self._yOriginSB.setSingleStep(10)
        self._yOriginSB.setPrefix("y: ")
        self._yOriginSB.setMaximum(10000)
        self.layout().addWidget(self._yOriginSB, 1, 2)

        # Signal / Slot connection
        self._widthSB.editingFinished.connect(self.__roiChanged)
        self._heightSB.editingFinished.connect(self.__roiChanged)
        self._xOriginSB.editingFinished.connect(self.__roiChanged)
        self._yOriginSB.editingFinished.connect(self.__roiChanged)

    def __roiChanged(self, *args, **kwargs):
        self.sigRoiChanged.emit((self.getROIDims(), self.getROIOrigin()))

    def setLimits(self, width, height):
        """

        :param x: width maximum value
        :param height: height maximum value
        """
        for spinButton in (self._widthSB, self._heightSB):
            spinButton.blockSignals(True)
        assert type(width) is int
        assert type(height) is int
        valueChanged = False
        if self._widthSB.value() > width:
            self._widthSB.setValue(width)
            valueChanged = True
        if self._heightSB.value() > height:
            self._heightSB.setValue(height)
            valueChanged = True

        # if this is the first limit definition, propose default width and
        # height
        if self._widthSB.value() == 0:
            self._widthSB.setValue(min(256, width))
            valueChanged = True
        if self._heightSB.value() == 0:
            self._heightSB.setValue(min(256, height))
            valueChanged = True

        # define minimum / maximum
        self._widthSB.setRange(1, width)
        self._heightSB.setRange(1, height)
        for spinButton in (self._widthSB, self._heightSB):
            spinButton.blockSignals(False)
        if valueChanged is True:
            self.__roiChanged()

    def getROIDims(self) -> tuple | None:
        """

        :return: (width, height) or None
        """
        if self.isEnabled():
            return (self._widthSB.value(), self._heightSB.value())
        else:
            return None

    def getROIOrigin(self):
        return (self._xOriginSB.value(), self._yOriginSB.value())

    def setEnabled(self, *arg, **kwargs):
        qt.QWidget.setEnabled(self, *arg, **kwargs)
        self.__roiChanged()

    def setScan(self, scan):
        if not self._already_set:
            self._already_set = True
            try:
                x_origin = scan.dim_1 // 2
                y_origin = scan.dim_2 // 2
                self._xOriginSB.setValue(x_origin)
                self._yOriginSB.setValue(y_origin)
            except Exception:
                _logger.warning(f"unable to determine origin for {scan}")


class AxisSettingsTabWidget(qt.QTabWidget):
    """
    TabWidget containing all the settings of the AxisTask
    """

    sigLockModeChanged = qt.Signal(bool)
    """signal emitted when the mode lock change"""

    sigUrlChanged = qt.Signal()
    """Signal emit when frames urls changed"""

    sigModeChanged = qt.Signal(str)
    """Signal emit when mode (algorithm) is changed"""

    sigUpdateXRotAxisPixelPosOnNewScan = qt.Signal()

    def __init__(
        self,
        recons_params: QAxisRP | None,
        parent=None,
    ):
        """

        :param recons_params: reconstruction parameters edited by the widget
        """
        qt.QTabWidget.__init__(self, parent)
        assert recons_params is not None
        # first tab 'calculation widget'
        self._calculationWidget = CalculationWidget(
            parent=self, axis_params=recons_params
        )

        # second tab: options
        self._optionsWidget = AxisOptionsWidget(parent=self, axis_params=recons_params)
        self._inputWidget = InputWidget(parent=self, axis_params=recons_params)

        for widget in self._calculationWidget, self._optionsWidget:
            spacer = qt.QWidget(self)
            spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
            widget.layout().addWidget(spacer)

        self.addTab(self._calculationWidget, "calculation")
        self.addTab(self._optionsWidget, "options")
        # simplify set up. Hide options
        self.addTab(self._inputWidget, "input")

        # set up
        self.setAxisParams(recons_params)
        self._updatePossibleInput()

        # connect signal / slot
        self._calculationWidget.sigLockModeChanged.connect(self.sigLockModeChanged)
        self.sigModeChanged.connect(self._updatePossibleInput)
        self.sigModeChanged.connect(self._updatePossibleOptions)
        self._inputWidget._sigUrlChanged.connect(self._urlChanged)
        self._calculationWidget.sigModeChanged.connect(self.sigModeChanged)
        self._calculationWidget.sigUpdateXRotAxisPixelPosOnNewScan.connect(
            self.sigUpdateXRotAxisPixelPosOnNewScan
        )

    def _urlChanged(self):
        self.sigUrlChanged.emit()

    def setScan(self, scan):
        if scan is not None:
            self._calculationWidget.setScan(scan)
            self._inputWidget.setScan(scan)

    def setAxisParams(self, axis):
        with block_signals(self):
            self._calculationWidget.setAxisParams(axis)
            self._optionsWidget.setAxisParams(axis)
            self._inputWidget.setAxisParams(axis)

    def _updatePossibleInput(self):
        """Update Input tab according to the current mode"""
        current_mode = self.getMode()
        valid_inputs = axis_mode.AXIS_MODE_METADATAS[current_mode].valid_inputs
        if valid_inputs is None:
            self._inputWidget.setEnabled(False)
        else:
            self._inputWidget.setEnabled(True)
            self._inputWidget.setValidInputs(valid_inputs)

    def _updatePossibleOptions(self):
        mode = self.getMode()
        self._optionsWidget.setMode(mode)

    # expose API
    def isModeLock(self) -> bool:
        return self._calculationWidget.isModeLock()

    def setModeLock(self, mode):
        self._calculationWidget.setModeLock(mode=mode)

    def setEstimatedCorValue(self, value):
        self._calculationWidget.setEstimatedCorValue(value=value)

    def getEstimatedCor(self):
        return self._calculationWidget.getEstimatedCor()

    def updateXRotationAxisPixelPositionOnNewScan(self) -> bool:
        return self._calculationWidget.updateXRotationAxisPixelPositionOnNewScan()

    def setUpdateXRotationAxisPixelPositionOnNewScan(self, update: bool):
        self._calculationWidget.setUpdateXRotationAxisPixelPositionOnNewScan(
            update=update
        )

    def getMode(self):
        """Return algorithm to use for axis calculation"""
        return self._calculationWidget.getMode()

    def isYAxisInverted(self) -> bool:
        return self._calculationWidget._estimatedCorWidget.isYAxisInverted()

    def setYAxisInverted(self, checked: bool):
        return self._calculationWidget._estimatedCorWidget.setYAxisInverted(
            checked=checked
        )


class _ShiftControl(qt.QWidget):
    """
    Widget to control the shift step we want to apply
    """

    sigShiftLeft = qt.Signal()
    """Signal emitted when the left button is activated"""
    sigShiftRight = qt.Signal()
    """Signal emitted when the right button is activated"""
    sigShiftTop = qt.Signal()
    """Signal emitted when the top button is activated"""
    sigShiftBottom = qt.Signal()
    """Signal emitted when the bottom button is activated"""
    sigReset = qt.Signal()
    """Signal emitted when the reset button is activated"""
    sigAuto = qt.Signal()
    """Signal emitted when the auto button is activated"""
    sigShiftChanged = qt.Signal(float, float)
    """Signal emitted ony when xLE and yLE edition is finished"""

    def __init__(self, parent, shift_mode: ShiftMode):
        """

        :param parent: qt.QWidget
        :param shift_mode: what are the shift we want to control
        """
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QGridLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

        self._leftButton = qt.QPushButton("left", parent=self)
        self.layout().addWidget(self._leftButton, 1, 0)

        self._rightButton = qt.QPushButton("right", parent=self)
        self.layout().addWidget(self._rightButton, 1, 3)

        self._shiftInfo = _ShiftInformation(parent=self)
        self.layout().addWidget(self._shiftInfo, 1, 1)
        self._shiftInfo._updateShiftInfo(x=0.0, y=0.0)

        self._topButton = qt.QPushButton("top", parent=self)
        self.layout().addWidget(self._topButton, 0, 1)

        self._bottomButton = qt.QPushButton("bottom", parent=self)
        self.layout().addWidget(self._bottomButton, 2, 1)

        self._resetButton = qt.QPushButton("reset", parent=self)
        self.layout().addWidget(self._resetButton, 3, 2, 3, 4)

        self._autoButton = qt.QPushButton("auto", parent=self)
        self.layout().addWidget(self._autoButton, 3, 0, 3, 2)
        self._autoButton.hide()

        # Signal / Slot connection
        self._leftButton.pressed.connect(self.sigShiftLeft.emit)
        self._rightButton.pressed.connect(self.sigShiftRight.emit)
        self._topButton.pressed.connect(self.sigShiftTop.emit)
        self._bottomButton.pressed.connect(self.sigShiftBottom.emit)
        self._resetButton.pressed.connect(self.sigReset.emit)
        self._autoButton.pressed.connect(self.sigAuto.emit)
        self._shiftInfo.sigShiftChanged.connect(self.sigShiftChanged.emit)

        # expose API
        self._updateShiftInfo = self._shiftInfo._updateShiftInfo

        self.setShiftMode(shift_mode)

    def setShiftMode(self, shift_mode):
        show_x_shift = shift_mode in (ShiftMode.x_only, ShiftMode.x_and_y)
        show_y_shift = shift_mode in (ShiftMode.y_only, ShiftMode.x_and_y)
        self._leftButton.setVisible(show_x_shift)
        self._rightButton.setVisible(show_x_shift)
        self._topButton.setVisible(show_y_shift)
        self._bottomButton.setVisible(show_y_shift)
        self._shiftInfo._xLE.setVisible(show_x_shift)
        self._shiftInfo._xLabel.setVisible(show_x_shift)
        self._shiftInfo._yLE.setVisible(show_y_shift)
        self._shiftInfo._yLabel.setVisible(show_y_shift)


class _ImgOpts(qt.QGroupBox):
    sigSubSamplingChanged = qt.Signal()
    """Signal emitted when the sub sampling change"""

    def __init__(self, parent, title="Image Option"):
        super().__init__(title, parent)
        self.setLayout(qt.QFormLayout())
        self._subSamplingQSpinBox = qt.QSpinBox(self)
        self.layout().addRow("sub-sampling:", self._subSamplingQSpinBox)
        self._subSamplingQSpinBox.setMinimum(1)

        # set up
        self._subSamplingQSpinBox.setValue(1)

        # connect signal / slot
        self._subSamplingQSpinBox.valueChanged.connect(self._subSamplingChanged)

    def _subSamplingChanged(self):
        self.sigSubSamplingChanged.emit()

    def getSubSampling(self):
        return self._subSamplingQSpinBox.value()

    def setSubSampling(self, value):
        return self._subSamplingQSpinBox.setValue(int(value))


class _ShiftInformation(qt.QWidget):
    """
    Widget displaying information about the current x and y shift.
    Both x shift and y shift are editable.
    """

    class _ShiftLineEdit(qt.QLineEdit):
        def __init__(self, *args, **kwargs):
            qt.QLineEdit.__init__(self, *args, **kwargs)
            self._defaultBackgroundColor = None
            # validator
            validator = qt.QDoubleValidator(parent=self, decimals=2)
            self.setValidator(validator)
            self._getDefaultBackgroundColor()
            # connect signal / slot
            self.textEdited.connect(self._userEditing)
            self.editingFinished.connect(self._userEndEditing)

        def sizeHint(self):
            return qt.QSize(40, 10)

        def _getDefaultBackgroundColor(self):
            if self._defaultBackgroundColor is None:
                self._defaultBackgroundColor = self.palette().color(
                    self.backgroundRole()
                )
            return self._defaultBackgroundColor

        def _userEditing(self, *args, **kwargs):
            palette = self.palette()
            palette.setColor(self.backgroundRole(), EDITING_BACKGROUND_COLOR)
            self.setPalette(palette)

        def _userEndEditing(self, *args, **kwargs):
            palette = self.palette()
            palette.setColor(
                self.backgroundRole(),
                self._getDefaultBackgroundColor(),
            )
            self.setPalette(palette)

    sigShiftChanged = qt.Signal(float, float)
    """Signal emitted ony when xLE and yLE edition is finished"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        self._xLabel = qt.QLabel("x=", parent=self)
        self.layout().addWidget(self._xLabel)
        self._xLE = _ShiftInformation._ShiftLineEdit("", parent=self)
        self.layout().addWidget(self._xLE)

        self._yLabel = qt.QLabel("y=", parent=self)
        self.layout().addWidget(self._yLabel)
        self._yLE = _ShiftInformation._ShiftLineEdit("", parent=self)
        self.layout().addWidget(self._yLE)

        # connect Signal / Slot
        self._xLE.editingFinished.connect(self._shiftChanged)
        self._yLE.editingFinished.connect(self._shiftChanged)

    def _updateShiftInfo(self, x, y):
        with block_signals(self):
            if x is None:
                x = 0.0
            if y is None:
                y = 0.0
            x_text = x
            if x_text != "...":
                x_text = "%.1f" % float(x)
            y_text = y
            if y_text != "...":
                y_text = "%.1f" % float(y)
            self._xLE.setText(x_text)
            self._yLE.setText(y_text)

    def _shiftChanged(self, *args, **kwargs):
        self.sigShiftChanged.emit(float(self._xLE.text()), float(self._yLE.text()))



from __future__ import annotations

from silx.gui.plot import Plot2D
from tomwer.gui.utils.buttons import TapeMeasureToolButton
from tomwer.gui.settings import Y_AXIS_DOWNWARD


class TomoPlot2D(Plot2D):
    """Plot set up for tomography and to display reconstructed slices"""

    def __init__(self, parent=None, backend=None):
        super().__init__(parent, backend)
        self.setYAxisInverted(Y_AXIS_DOWNWARD)
        self.setKeepDataAspectRatio(True)
        self.setAxesDisplayed(False)

        self._tapeMeasureButton = TapeMeasureToolButton(parent=self, plot=self)
        self._tapeMeasureButton.setCheckable(True)

        self.toolBar().addWidget(self._tapeMeasureButton)

    def setVoxelSize(self, size: float | None):
        """set the pixel size to be used by the ruler"""
        self._tapeMeasureButton.setPixelSize(pixel_size=size)

    def setToolbarVisible(self, visible):
        self.toolBar().setVisible(visible)



# coding: utf-8
from __future__ import annotations


import os

import pytest
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from tomwer.gui.reconstruction.nabu.nabuconfig.output import _NabuOutputConfig
from tomwer.gui.reconstruction.nabu.nabuconfig.phase import _NabuPhaseConfig
from tomwer.gui.reconstruction.nabu.nabuconfig.preprocessing import (
    _NabuPreProcessingConfig,
    RingCorrectionMethod,
)
from tomwer.gui.reconstruction.nabu.nabuconfig.reconstruction import (
    _NabuReconstructionConfig,
)
from tomwer.gui.reconstruction.nabu.nabuflow import NabuFlowControl
from tomwer.gui.reconstruction.nabu.volume import NabuVolumeTabWidget
from tomwer.gui.reconstruction.axis.EstimatedCORWidget import EstimatedCORWidget
from tomwer.tests.utils import skip_gui_test
from tomwer.core.process.output import ProcessDataOutputDirMode
from tomwer.tests.conftest import qtapp  # noqa F401
from tomwer.synctools.axis import QAxisRP


class ProcessClass:
    """Simple class for unit tests"""

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestNabuFlow(TestCaseQt):
    def setUp(self) -> None:
        TestCaseQt.setUp(self)
        self.nabuWidget = NabuFlowControl(parent=None, direction="vertical")

    def tearDown(self):
        self.nabuWidget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.nabuWidget.close()
        del self.nabuWidget

    def testFlow1(self):
        style = qt.QApplication.style()
        icon_1 = style.standardIcon(qt.QStyle.SP_DialogApplyButton)
        icon_2 = style.standardIcon(qt.QStyle.SP_FileLinkIcon)
        icon_3 = style.standardIcon(qt.QStyle.SP_ArrowLeft)
        icon_4 = style.standardIcon(qt.QStyle.SP_ArrowRight)
        icon_5 = style.standardIcon(qt.QStyle.SP_BrowserStop)

        preprocess = "reading files", ProcessClass(name="other preprocessing")
        preprocess_icons = None, icon_1
        self.nabuWidget.setPreProcessing(processes=preprocess, icons=preprocess_icons)

        processes = (
            "processing 1",
            ProcessClass("in between processing"),
            "other processing",
        )
        processes_icons = icon_2, None, icon_3
        self.nabuWidget.setProcessing(processes=processes, icons=processes_icons)

        postprocess = "post processing", ProcessClass("writing result")
        postprocess_icons = icon_4, icon_5
        self.nabuWidget.setPostProcessing(
            processes=postprocess, icons=postprocess_icons
        )


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestNabuPreProcConfig(TestCaseQt):
    def setUp(self):
        TestCaseQt.setUp(self)
        self.nabuWidget = _NabuPreProcessingConfig(parent=None)

    def tearDown(self):
        self.nabuWidget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.nabuWidget.close()
        del self.nabuWidget
        self.qapp.processEvents()

    def testGetInitialConfiguration(self):
        """Test that the get configuration is working"""
        ini_conf = {
            "flatfield": 1,
            "double_flatfield": 0,
            "dff_sigma": 0.0,
            "ccd_filter_enabled": 0,
            "ccd_filter_threshold": 0.04,
            "log_min_clip": 1e-6,
            "log_max_clip": 10.0,
            "take_logarithm": True,
            "normalize_srcurrent": 0,
            "sino_rings_correction": RingCorrectionMethod.NONE.value,
            "sino_rings_options": "sigma=1.0 ; levels=10 ; padding=False",
            "tilt_correction": "",
            "autotilt_options": "",
            "rotate_projections_center": "",
        }
        self.assertEqual(self.nabuWidget.getConfiguration(), ini_conf)

    def testSetConfiguration(self):
        """Test that the set configuration is working"""
        conf = {
            "flatfield": 0,
            "double_flatfield": 1,
            "dff_sigma": 2.0,
            "ccd_filter_enabled": 1,
            "ccd_filter_threshold": 0.98,
            "log_min_clip": 1e-3,
            "log_max_clip": 250.0,
            "take_logarithm": False,
            "normalize_srcurrent": 1,
            "sino_rings_correction": RingCorrectionMethod.MUNCH.value,
            "sino_rings_options": "sigma=1.4 ; levels=11 ; padding=True",
            "tilt_correction": "1d-correlation",
            "autotilt_options": "low_pass=1; high_pass=20",
            "rotate_projections_center": (2.1, 3.0),
        }
        self.nabuWidget.setConfiguration(conf=conf)
        self.assertEqual(self.nabuWidget.getConfiguration(), conf)


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestNabuPhaseConfig(TestCaseQt):
    def setUp(self):
        TestCaseQt.setUp(self)
        self.nabuWidget = _NabuPhaseConfig(parent=None)

    def tearDown(self):
        self.nabuWidget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.nabuWidget.close()
        del self.nabuWidget

    def testGetInitialConfiguration(self):
        """Test that the get configuration is working"""
        ini_conf = {
            "method": "Paganin",
            "delta_beta": "100.0",
            "padding_type": "edge",
            "unsharp_coeff": 0,
            "unsharp_sigma": 0,
            "beam_shape": "parallel",
            "ctf_advanced_params": " length_scale=1e-05; lim1=1e-05; lim2=0.2; normalize_by_mean=True",
            "ctf_geometry": " z1_v=None; z1_h=None; detec_pixel_size=None; magnification=True",
        }
        self.assertEqual(self.nabuWidget.getConfiguration(), ini_conf)

    def testSetConfiguration(self):
        """Test that the set configuration is working"""
        conf = {
            "method": "Paganin",
            "delta_beta": "200.0",
            "padding_type": "zeros",
            "unsharp_coeff": 3.6,
            "unsharp_sigma": 2.1,
            "beam_shape": "cone",
            "ctf_advanced_params": " length_scale=1e-05; lim1=1e-05; lim2=0.2; normalize_by_mean=True",
            "ctf_geometry": " z1_v=0.0; z1_h=0.0; detec_pixel_size=None; magnification=True",
        }
        self.nabuWidget.setConfiguration(conf)
        self.nabuWidget.show()
        # check visibility of some widgets
        unsharp_widget = self.nabuWidget._unsharpOpts
        self.assertEqual(unsharp_widget._unsharpCoeffQLE.text(), "3.6")
        self.assertTrue(unsharp_widget._unsharpCoeffCB.isChecked())

        self.assertEqual(unsharp_widget._unsharpSigmaQLE.text(), "2.1")
        self.assertTrue(unsharp_widget._unsharpSigmaCB.isChecked())
        paganin_widget = self.nabuWidget._paganinOpts
        self.assertEqual(paganin_widget._deltaBetaQLE.text(), "200.0")
        self.assertEqual(paganin_widget._paddingTypeCB.currentText(), "zeros")

        # check the generated configuration
        self.assertEqual(self.nabuWidget.getConfiguration(), conf)


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestNabuReconstructionConfig(TestCaseQt):
    def setUp(self):
        TestCaseQt.setUp(self)
        self.nabuWidget = _NabuReconstructionConfig(parent=None)

    def tearDown(self):
        self.nabuWidget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.nabuWidget.close()
        del self.nabuWidget

    def testGetInitialConfiguration(self):
        """Test that the get configuration is working"""
        ini_conf = {
            "method": "FBP",
            "slice_plane": "XY",
            "angles_file": "",
            "axis_correction_file": "",
            "angle_offset": 0.0,
            "fbp_filter_type": "ramlak",
            "padding_type": "zeros",
            "start_x": 0,
            "end_x": -1,
            "start_y": 0,
            "end_y": -1,
            "start_z": 0,
            "end_z": -1,
            "iterations": 200,
            "optim_algorithm": "chambolle-pock",
            "weight_tv": 1.0e-2,
            "preconditioning_filter": 1,
            "rotation_axis_position": "",
            "positivity_constraint": 1,
            "translation_movements_file": "",
            "clip_outer_circle": 0,
            "centered_axis": 0,
        }
        self.assertEqual(self.nabuWidget.getConfiguration(), ini_conf)

    def testSetConfiguration(self):
        """Test that the set configuration is working"""
        ini_conf = {
            "method": "FBP",
            "slice_plane": "XZ",
            "angles_file": "",
            "axis_correction_file": "",
            "angle_offset": 12.5,
            "fbp_filter_type": "none",
            "padding_type": "edges",
            "start_x": 0,
            "end_x": 23,
            "start_y": 12,
            "end_y": 56,
            "start_z": 560,
            "end_z": -1,
            "iterations": 20,
            "optim_algorithm": "chambolle-pock",
            "weight_tv": 1.5e-2,
            "preconditioning_filter": 0,
            "rotation_axis_position": "",
            "positivity_constraint": 0,
            "translation_movements_file": "my_file.csv",
            "clip_outer_circle": 1,
            "centered_axis": 1,
        }
        self.nabuWidget.setConfiguration(ini_conf)
        self.qapp.processEvents()
        self.nabuWidget.show()

        # check visibility of some widgets
        self.assertEqual(self.nabuWidget._angleOffsetQDSB.value(), 12.5)
        subRegionWidget = self.nabuWidget._subRegionSelector
        self.assertFalse(subRegionWidget._xSubRegion._minCB.isChecked())
        self.assertFalse(subRegionWidget._xSubRegion._minQLE.isEnabled())
        self.assertTrue(subRegionWidget._xSubRegion._maxQLE.isEnabled())
        self.assertTrue(subRegionWidget._ySubRegion._minQLE.isEnabled())
        self.assertTrue(subRegionWidget._ySubRegion._maxQLE.isEnabled())
        self.assertTrue(subRegionWidget._zSubRegion._minQLE.isEnabled())
        self.assertFalse(subRegionWidget._zSubRegion._maxQLE.isEnabled())
        self.assertFalse(self.nabuWidget._preconditioningFilter.isChecked())
        self.assertTrue(self.nabuWidget._clipOuterCircleCB.isChecked())
        self.assertTrue(self.nabuWidget._centeredAxisCB.isChecked())
        assert self.nabuWidget._axisQCB.currentText() == ini_conf["slice_plane"]

        # check the generated configuration
        self.assertEqual(self.nabuWidget.getConfiguration(), ini_conf)


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestNabuOutputConfig(TestCaseQt):
    """Test the output configuration interface"""

    def setUp(self):
        TestCaseQt.setUp(self)
        self.nabuWidget = _NabuOutputConfig(parent=None)

    def tearDown(self):
        self.nabuWidget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.nabuWidget.close()
        self.nabuWidget = None
        self.qapp.processEvents()

    def testGetConfiguration(self):
        ini_conf = {
            "file_format": "hdf5",
            "location": "",
            "output_dir_mode": "same folder as scan",
        }
        self.assertEqual(self.nabuWidget.getConfiguration(), ini_conf)

    def testSetConfiguration(self):
        conf = {
            "file_format": "tiff",
            "location": os.sep.join(("tmp", "my_output")),
            "output_dir_mode": "other",
        }
        self.nabuWidget.setConfiguration(conf)
        self.nabuWidget.show()
        self.qapp.processEvents()
        # check some widget visibility
        self.assertTrue(self.nabuWidget._output_dir_widget._outputDirQLE.isVisible())
        self.assertTrue(self.nabuWidget._output_dir_widget._otherDirRB.isChecked())

        self.assertEqual(self.nabuWidget.getConfiguration(), conf)


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestNabuVolumeWidget(TestCaseQt):
    def setUp(self):
        TestCaseQt.setUp(self)
        self.nabuWidget = NabuVolumeTabWidget(parent=None)
        self.nabuWidget.setConfigurationLevel("advanced")

    def tearDown(self):
        self.nabuWidget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.nabuWidget.close()
        self.nabuWidget = None

    def testGetConfiguration(self):
        ini_conf = {
            "start_z": 0,
            "end_z": -1,
            "gpu_mem_fraction": 0.9,
            "output_dir_mode": ProcessDataOutputDirMode.IN_SCAN_FOLDER.value,
            "overwrite_output_location": False,
            "postproc": {"output_histogram": 1},
            "cpu_mem_fraction": 0.9,
            "use_phase_margin": True,
            "new_output_file_format": "",
            "new_output_location": "",
        }
        self.assertEqual(self.nabuWidget.getConfiguration(), ini_conf)

    def testSetConfiguration(self):
        conf = {
            "start_z": 10,
            "end_z": 24,
            "gpu_mem_fraction": 0.8,
            "postproc": {"output_histogram": 0},
            "cpu_mem_fraction": 0.1,
            "use_phase_margin": False,
            "new_output_file_format": "hdf5",
            "new_output_location": "/new/location",
        }
        self.nabuWidget.setConfiguration(conf)
        self.qapp.processEvents()

        # update the config dict has setting an other custom output directory update some items
        conf.update(
            {
                "overwrite_output_location": True,
                "output_dir_mode": ProcessDataOutputDirMode.OTHER.value,
            }
        )
        self.assertEqual(self.nabuWidget.getConfiguration(), conf)


def test_EstimatedCorWidget(qtapp):  # noqa F811
    """test of EstimatedCorWidget"""
    EstimatedCORWidget(parent=None, axis_params=QAxisRP())



# coding: utf-8

"""
Contains processing relative to a `Slurmcluster`
"""

from __future__ import annotations

import logging
from tomwer.core.settings import SlurmSettings as _SlurmSettings

_logger = logging.getLogger(__name__)


class SlurmClusterConfiguration:
    """Object shipping the configuration of a slurm cluster"""

    def __init__(
        self,
        n_cpu_per_task=_SlurmSettings.N_CORES_PER_TASK,
        n_tasks=_SlurmSettings.N_TASKS,
        memory=_SlurmSettings.MEMORY_PER_WORKER,
        queue=_SlurmSettings.PARTITION,
        n_gpus=_SlurmSettings.N_GPUS_PER_WORKER,
        project_name=_SlurmSettings.PROJECT_NAME,
        walltime=_SlurmSettings.DEFAULT_WALLTIME,
        python_venv=_SlurmSettings.PYTHON_VENV,
        n_jobs=_SlurmSettings.N_JOBS,
        modules_to_load: tuple = _SlurmSettings.MODULES_TO_LOAD,
        sbatch_extra_params: dict = _SlurmSettings.SBATCH_EXTRA_PARAMS,
    ) -> None:
        self._n_cpu_per_task = n_cpu_per_task
        self._n_task = n_tasks
        self._memory = memory
        self._queue = queue
        self._n_gpus = n_gpus
        self._project_name = project_name
        self._walltime = walltime
        self._python_venv = python_venv
        self._modules_to_load = modules_to_load
        self._n_jobs = n_jobs
        self._sbatch_extra_params = sbatch_extra_params
        if python_venv not in (None, "") and len(modules_to_load) > 0:
            _logger.warning(
                "Either 'modules to load' of 'python venv' can be provided. Not both"
            )

    @property
    def n_cpu_per_task(self):
        return self._n_cpu_per_task

    @property
    def n_task(self):
        return self._n_task

    @property
    def memory(self):
        return self._memory

    @property
    def queue(self):
        return self._queue

    @property
    def n_gpus(self):
        return self._n_gpus

    @property
    def n_jobs(self):
        return self._n_jobs

    @property
    def project_name(self):
        return self._project_name

    @property
    def walltime(self):
        return self._walltime

    @property
    def python_venv(self):
        return self._python_venv

    @property
    def modules_to_load(self) -> tuple:
        return self._modules_to_load

    @property
    def port_range(self) -> tuple:
        """port range as (start:int, strop:int, step: int)"""
        return self._port_range

    @property
    def sbatch_extra_params(self) -> dict:
        return self._sbatch_extra_params

    @property
    def dashboard_port(self):
        return self._dashboard_port

    def to_dict(self) -> dict:
        return {
            "cpu-per-task": self.n_cpu_per_task,
            "n_tasks": self.n_task,
            "n_jobs": self.n_jobs,
            "memory": self.memory,
            "partition": self.queue,
            "n_gpus": self.n_gpus,
            "job_name": self.project_name,
            "walltime": self.walltime,
            "python_venv": self.python_venv,
            "modules": self.modules_to_load,
            "sbatch_extra_params": self.sbatch_extra_params,
        }

    @staticmethod
    def from_dict(dict_: dict):
        return SlurmClusterConfiguration().load_from_dict(dict_=dict_)

    def load_from_dict(self, dict_: dict):
        if "cpu-per-task" in dict_:
            self._n_cpu_per_task = dict_["cpu-per-task"]
        if "n_tasks" in dict_:
            self._n_task = dict_["n_tasks"]
        if "n_jobs" in dict_:
            self._n_jobs = dict_["n_jobs"]
        if "memory" in dict_:
            self._memory = dict_["memory"]
        if "partition" in dict_:
            self._queue = dict_["partition"]
        if "n_gpus" in dict_:
            self._n_gpus = dict_["n_gpus"]
        if "job_name" in dict_:
            self._project_name = dict_["job_name"]
        if "walltime" in dict_:
            self._walltime = dict_["walltime"]
        if "python_venv" in dict_:
            self._python_venv = dict_["python_venv"]
        modules = dict_.get("modules", None)
        if modules is not None:
            self._modules_to_load = modules
        if "port_range" in dict_:
            self._port_range = dict_["port_range"]
        if "dashboard_port" in dict_:
            self._dashboard_port = dict_["dashboard_port"]
        if "sbatch_extra_params" in dict_:
            self._sbatch_extra_params = dict_["sbatch_extra_params"]
        return self



from __future__ import annotations

import logging

import sys
from silx.gui import qt

from tomwer.core.process.reconstruction.nabu.utils import (
    _NabuStages,
    RingCorrectionMethod,
)
from tomwer.gui.reconstruction.nabu.nabuconfig.base import _NabuStageConfigBase
from tomwer.gui.utils.scrollarea import QComboBoxIgnoreWheel as QComboBox
from tomwer.gui.utils.scrollarea import QDoubleSpinBoxIgnoreWheel as QDoubleSpinBox
from tomwer.gui.utils.scrollarea import QSpinBoxIgnoreWheel as QSpinBox
from tomwer.utils import docstring

_logger = logging.getLogger(__name__)


class _NabuPreProcessingConfig(_NabuStageConfigBase, qt.QWidget):
    """
    Widget to define the configuration of the nabu preprocessing
    """

    sigConfChanged = qt.Signal(str)
    """Signal emitted when the configuration change. Parameter is the option
    modified
    """

    def __init__(self, parent):
        _NabuStageConfigBase.__init__(self, stage=_NabuStages.PRE)
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QGridLayout())

        # default options
        ## flat field
        self._flatFieldCB = qt.QCheckBox("flat field correction", self)
        self._flatFieldCB.setToolTip("Whether to enable flat-field " "normalization")
        self.layout().addWidget(self._flatFieldCB, 0, 0, 1, 2)
        self.registerWidget(self._flatFieldCB, "optional")

        ## double flat field
        self._dffCB = qt.QCheckBox("double flat field correction", self)
        self._dffCB.setToolTip("Whether to enable double flat field " "normalization")
        self.layout().addWidget(self._dffCB, 1, 0, 1, 2)
        self._dffSigmaLabel = qt.QLabel("sigma:", self)
        self._dffSigmaLabel.setAlignment(qt.Qt.AlignRight)
        self.layout().addWidget(self._dffSigmaLabel, 1, 2, 1, 1)
        self._dffSigmaQDSB = QDoubleSpinBox(parent=self)
        self._dffSigmaQDSB.setMinimum(0.0)
        self._dffSigmaQDSB.setDecimals(2)
        self._dffSigmaQDSB.setSingleStep(0.1)
        self._dffSigmaQDSB.setToolTip(
            "Sigma value to give to the double flat field unsharp mask"
        )
        self.layout().addWidget(self._dffSigmaQDSB, 1, 3, 1, 1)
        self.registerWidget(self._flatFieldCB, "required")
        self._dffOptWidgets = [
            self.registerWidget(self._dffSigmaLabel, "required"),
            self.registerWidget(self._dffSigmaQDSB, "required"),
        ]

        ## sinogram ring correction
        self._sinoRingCorrectionCB = qt.QLabel("rings removal method", self)
        self._sinoRingCorrectionCB.setToolTip("Sinogram rings removal method")
        self.layout().addWidget(self._sinoRingCorrectionCB, 2, 0, 1, 2)
        self.registerWidget(self._sinoRingCorrectionCB, "required")

        self._sinoRingCorrectionMthd = QComboBox(parent=self)
        for method in RingCorrectionMethod:
            self._sinoRingCorrectionMthd.addItem(method.value)
        ## force method to be None by default
        idx = self._sinoRingCorrectionMthd.findText(RingCorrectionMethod.NONE.value)
        self._sinoRingCorrectionMthd.setCurrentIndex(idx)

        self.layout().addWidget(self._sinoRingCorrectionMthd, 2, 2, 1, 1)
        self.registerWidget(self._sinoRingCorrectionMthd, "required")

        self._sinoRingsOpts = SinoRingsOptions(parent=self)
        self.layout().addWidget(self._sinoRingsOpts, 3, 1, 1, 3)

        ## ccd filter
        self._ccdFilterCB = qt.QCheckBox("CCD hot spot correction", self)
        self._ccdFilterCB.setToolTip("Whether to enable the CCD hotspots " "correction")
        self.layout().addWidget(self._ccdFilterCB, 4, 0, 1, 2)
        self.registerWidget(self._ccdFilterCB, "optional")

        ## ccd filter threshold
        self._ccdHotspotLabel = qt.QLabel("threshold:", self)
        self._ccdHotspotLabel.setAlignment(qt.Qt.AlignRight)
        self.layout().addWidget(self._ccdHotspotLabel, 5, 2, 1, 1)
        self._ccdThreshold = QDoubleSpinBox(self)
        self._ccdThreshold.setMinimum(0.0)
        self._ccdThreshold.setMaximum(999999)
        self._ccdThreshold.setSingleStep(0.01)
        self._ccdThreshold.setDecimals(6)
        tooltip = (
            "If ccd_filter_enabled = 1, a median filter is applied on "
            "the 3X3 neighborhood\nof every pixel. If a pixel value "
            "exceeds the median value more than this parameter,\nthen "
            "the pixel value is replaced with the median value."
        )
        self._ccdThreshold.setToolTip(tooltip)
        self.layout().addWidget(self._ccdThreshold, 5, 3, 1, 1)
        self._ccdOptWidgets = [
            self.registerWidget(self._ccdHotspotLabel, "optional"),
            self.registerWidget(self._ccdThreshold, "optional"),
        ]

        ## sr current normalization
        self._normalizeCurrent = qt.QCheckBox(
            "Normalize by machine electric current", self
        )
        self._normalizeCurrent.setToolTip(
            "Whether to normalize frames with Synchrotron Current. This can correct the effect of a beam refill not taken into account by flats."
        )
        self.layout().addWidget(self._normalizeCurrent, 6, 0, 1, 2)
        self.registerWidget(self._normalizeCurrent, "required")

        ## take logarithm
        self._takeLogarithmCB = qt.QCheckBox("take logarithm", self)
        self.layout().addWidget(self._takeLogarithmCB, 7, 0, 1, 2)
        self.registerWidget(self._takeLogarithmCB, "advanced")

        ## log min clip value
        self._clipMinLogValueLabel = qt.QLabel("log min clip value:", self)
        self._clipMinLogValueLabel.setAlignment(qt.Qt.AlignRight)
        self.layout().addWidget(self._clipMinLogValueLabel, 8, 2, 1, 1)
        self._clipMinLogValue = QDoubleSpinBox(self)
        self._clipMinLogValue.setMinimum(0.0)
        self._clipMinLogValue.setMaximum(9999999)
        self._clipMinLogValue.setSingleStep(0.01)
        self._clipMinLogValue.setDecimals(6)
        self.layout().addWidget(self._clipMinLogValue, 8, 3, 1, 1)
        self._takeLogOpt = [
            self.registerWidget(self._clipMinLogValueLabel, "optional"),
            self.registerWidget(self._clipMinLogValue, "optional"),
        ]

        ## log max clip value
        self._clipMaxLogValueLabel = qt.QLabel("log max clip value:", self)
        self._clipMaxLogValueLabel.setAlignment(qt.Qt.AlignRight)
        self.layout().addWidget(self._clipMaxLogValueLabel, 9, 2, 1, 1)
        self._clipMaxLogValue = QDoubleSpinBox(self)
        self._clipMaxLogValue.setMinimum(0.0)
        self._clipMaxLogValue.setMaximum(9999999)
        self._clipMaxLogValue.setSingleStep(0.01)
        self._clipMaxLogValue.setDecimals(6)
        self.layout().addWidget(self._clipMaxLogValue, 9, 3, 1, 1)
        self._takeLogOpt.extend(
            [
                self.registerWidget(self._clipMaxLogValueLabel, "optional"),
                self.registerWidget(self._clipMaxLogValue, "optional"),
            ]
        )

        ## tilt correction
        self._tiltCorrection = TiltCorrection("tilt correction", self)
        self.registerWidget(self._tiltCorrection, "advanced")
        self.layout().addWidget(self._tiltCorrection, 10, 0, 1, 4)

        # option dedicated to Helical
        ## process file
        self._processFileLabel = qt.QLabel("file containing weights maps", self)
        self.registerWidget(self._processFileLabel, "advanced")
        self.layout().addWidget(self._processFileLabel, 20, 0, 1, 1)
        self._processFileQLE = qt.QLineEdit("", self)
        self.registerWidget(self._processFileQLE, "advanced")
        self._processFileQLE.setToolTip(
            "also know as 'process_file'. If you don't have this file it can be created from the 'helical-prepare-weights' widget"
        )
        self.layout().addWidget(self._processFileQLE, 20, 1, 1, 3)

        # style

        # spacer for style
        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer, 99, 0, 1, 1)

        # set up
        self._flatFieldCB.setChecked(True)
        self.setDFFOptVisible(False)

        self._ccdFilterCB.setChecked(False)
        self._normalizeCurrent.setChecked(False)
        self._ccdThreshold.setValue(0.04)

        self._clipMinLogValue.setValue(1e-6)
        self._clipMaxLogValue.setValue(10.0)
        self._takeLogarithmCB.setChecked(True)
        self.setCCDOptsVisible(False)
        self._sinoRingCorrectionMthd.setCurrentText("None")
        self._sinoRingsOpts.setVisible(False)
        self._tiltCorrection.setChecked(False)

        # connect signal / slot
        self._ccdFilterCB.toggled.connect(self.setCCDOptsVisible)
        self._takeLogarithmCB.toggled.connect(self.setLogClipValueVisible)
        self._flatFieldCB.toggled.connect(self._flatFieldChanged)
        self._dffCB.toggled.connect(self._dffChanged)
        self._dffCB.toggled.connect(self.setDFFOptVisible)
        self._dffSigmaQDSB.valueChanged.connect(self._dffSigmaChanged)
        self._ccdFilterCB.toggled.connect(self._ccdFilterChanged)
        self._normalizeCurrent.toggled.connect(self._normalizeCurrentChanged)
        self._ccdThreshold.editingFinished.connect(self._ccdFilterThresholdChanged)
        self._clipMinLogValue.editingFinished.connect(self._logMinClipChanged)
        self._clipMaxLogValue.editingFinished.connect(self._logMaxClipChanged)
        self._takeLogarithmCB.toggled.connect(self._takeLogarithmChanged)
        self._sinoRingCorrectionMthd.currentIndexChanged.connect(
            self._sinoRingCorrectionChanged
        )
        self._sinoRingsOpts._levelsMunch.valueChanged.connect(self._sinoRingOptsChanged)
        self._sinoRingsOpts._sigmaMunch.valueChanged.connect(self._sinoRingOptsChanged)

        self._tiltCorrection.toggled.connect(self._tiltCorrectionChanged)
        self._tiltCorrection.sigChanged.connect(self._tiltCorrectionChanged)

    def _flatFieldChanged(self, *args, **kwargs):
        self._signalConfChanged("flatfield")

    def _dffChanged(self, *args, **kwargs):
        self._signalConfChanged("double_flatfield")

    def _dffSigmaChanged(self, *args, **kwargs):
        self._signalConfChanged("dff_sigma")

    def _ccdFilterChanged(self, *args, **kwargs):
        self._signalConfChanged("ccd_filter_enabled")

    def _normalizeCurrentChanged(self, *args, **kwargs):
        self._signalConfChanged("normalize_srcurrent")

    def _ccdFilterThresholdChanged(self, *args, **kwargs):
        self._signalConfChanged("ccd_filter_threshold")

    def _logMinClipChanged(self, *args, **kwargs):
        self._signalConfChanged("log_min_clip")

    def _logMaxClipChanged(self, *args, **kwargs):
        self._signalConfChanged("log_max_clip")

    def _takeLogarithmChanged(self, *args, **kwargs):
        self._signalConfChanged("take_logarithm")

    def _sinoRingCorrectionChanged(self, *args, **kwargs):
        method = self.getSinoRingcorrectionMethod()
        if method is not RingCorrectionMethod.NONE.value:
            self._sinoRingsOpts.setVisible(True)
            self._sinoRingsOpts.setMethod(method)
        else:
            self._sinoRingsOpts.setVisible(False)

        self._signalConfChanged("sino_rings_correction")

    def _sinoRingOptsChanged(self, *args, **kwargs):
        self._signalConfChanged("sino_rings_options")

    def _tiltCorrectionChanged(self, *args, **kwargs):
        self._signalConfChanged("tilt_correction")

    def _signalConfChanged(self, param, *args, **kwargs):
        self.sigConfChanged.emit(param)

    def setDFFOptVisible(self, visible):
        for widget in self._dffOptWidgets:
            widget.setVisible(visible)

    def setCCDOptsVisible(self, visible):
        for widget in self._ccdOptWidgets:
            widget.setVisible(visible)

    def setLogClipValueVisible(self, visible):
        for widget in self._takeLogOpt:
            widget.setVisible(visible)

    def isFlatFieldActivate(self):
        return self._flatFieldCB.isChecked()

    def isDoubleFlatFieldActivate(self):
        return self._dffCB.isChecked()

    def getDFFSigma(self) -> float:
        """

        :return: double flat field sigma
        """
        return self._dffSigmaQDSB.value()

    def isCCDFilterActivate(self):
        return self._ccdFilterCB.isChecked()

    def getCCDThreshold(self) -> float:
        return float(self._ccdThreshold.text())

    def getNormalizeCurrent(self) -> bool:
        return self._normalizeCurrent.isChecked()

    def setNormalizeCurrent(self, normalize: bool) -> None:
        self._normalizeCurrent.setChecked(normalize)

    def getLogMinClipValue(self) -> float:
        return float(self._clipMinLogValue.text())

    def getLogMaxClipValue(self) -> float:
        return float(self._clipMaxLogValue.text())

    def getTakeLogarithm(self):
        return self._takeLogarithmCB.isChecked()

    def getSinoRingcorrectionMethod(self) -> str:
        return self._sinoRingCorrectionMthd.currentText()

    def getSinoRingcorrectionOptions(self) -> str:
        return " ; ".join(
            [
                f"{key}={value}"
                for key, value in self._sinoRingsOpts.getOptions().items()
            ]
        )

    def setSinoRingcorrectionOptions(self, options: str) -> None:
        opt_as_dict = {}
        for opt in options.split(";"):
            opt = opt.replace(" ", "")
            if len(opt.split("=")) == 2:
                key, value = opt.split("=")
                opt_as_dict[key] = value
            else:
                _logger.info(f"ignore option {opt}. Invalid synthax")

        self._sinoRingsOpts.setOptions(opt_as_dict)

    @docstring(_NabuStageConfigBase)
    def getConfiguration(self):
        tilt_correction, autotilt_opts = self._tiltCorrection.getTiltCorrection()
        return {
            "flatfield": int(self.isFlatFieldActivate()),
            "double_flatfield": int(self.isDoubleFlatFieldActivate()),
            "dff_sigma": self.getDFFSigma(),
            "ccd_filter_enabled": int(self.isCCDFilterActivate()),
            "ccd_filter_threshold": self.getCCDThreshold(),
            "take_logarithm": self.getTakeLogarithm(),
            "log_min_clip": self.getLogMinClipValue(),
            "log_max_clip": self.getLogMaxClipValue(),
            "sino_rings_correction": self.getSinoRingcorrectionMethod(),
            "sino_rings_options": self.getSinoRingcorrectionOptions(),
            "tilt_correction": tilt_correction,
            "autotilt_options": autotilt_opts,
            "normalize_srcurrent": int(self.getNormalizeCurrent()),
            "rotate_projections_center": self._tiltCorrection.getRotateProjectionsCenter()
            or "",
        }

    @docstring(_NabuStageConfigBase)
    def setConfiguration(self, conf):
        try:
            self._setConfiguration(conf)
        except Exception as e:
            _logger.error(e)

    def _setConfiguration(self, conf: dict):
        ff = conf.get("flatfield", None)
        if ff is not None:
            self._flatFieldCB.setChecked(bool(ff))

        dff = conf.get("double_flatfield", None)
        if dff is not None:
            self._dffCB.setChecked(bool(dff))

        dff_sigma = conf.get("dff_sigma", None)
        if dff_sigma not in (None, "", "none"):
            self._dffSigmaQDSB.setValue(float(dff_sigma))

        ccd_filter = conf.get("ccd_filter_enabled", None)
        if ccd_filter not in (None, "", "none"):
            self._ccdFilterCB.setChecked(bool(ccd_filter))

        ccd_filter_threshold = conf.get("ccd_filter_threshold", None)
        if ccd_filter_threshold not in (None, "", "none"):
            self._ccdThreshold.setValue(float(ccd_filter_threshold))

        normalize_srcurrent = conf.get("normalize_srcurrent", None)
        if normalize_srcurrent is not None:
            self.setNormalizeCurrent(bool(normalize_srcurrent))

        take_logarithm = conf.get("take_logarithm", None)
        if take_logarithm not in (None, "", "none"):
            self._takeLogarithmCB.setChecked(bool(take_logarithm))

        clip_value = conf.get("log_min_clip", None)
        if clip_value not in (None, "", "none"):
            self._clipMinLogValue.setValue(float(clip_value))

        clip_value = conf.get("log_max_clip", None)
        if clip_value not in (None, "", "none"):
            self._clipMaxLogValue.setValue(float(clip_value))

        sino_rings_correction = conf.get("sino_rings_correction", None)
        if sino_rings_correction is not None:
            if sino_rings_correction == "":
                sino_rings_correction = RingCorrectionMethod.NONE
            sino_rings_correction = RingCorrectionMethod(sino_rings_correction).value
            idx = self._sinoRingCorrectionMthd.findText(sino_rings_correction)
            if idx >= 0:
                self._sinoRingCorrectionMthd.setCurrentIndex(idx)
        sino_rings_options = conf.get("sino_rings_options", None)
        if sino_rings_options is not None:
            self.setSinoRingcorrectionOptions(options=sino_rings_options)

        tilt_correction = conf.get("tilt_correction")
        autotilt_options = conf.get("autotilt_options")
        self._tiltCorrection.setTiltCorrection(
            tilt_correction=tilt_correction, auto_tilt_options=autotilt_options
        )

        rotate_projections_center = conf.get("rotate_projections_center")
        if rotate_projections_center == "":
            rotate_projections_center = None
        self._tiltCorrection.setRotateProjectionsCenter(rotate_projections_center)


class SinoRingsOptions(qt.QWidget):
    _VO_DIMS = ("horizontaly", "horizontaly and vertically")

    def __init__(self, parent=None, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self._method = None
        self.setLayout(qt.QFormLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)
        # munch parameters
        self._sigmaMunchLabel = qt.QLabel("sigma", self)
        self._sigmaMunch = QDoubleSpinBox(self)
        self._sigmaMunch.setRange(0.0, 2147483647)
        self.layout().addRow(self._sigmaMunchLabel, self._sigmaMunch)

        self._levelsMunchLabel = qt.QLabel("levels", self)
        self._levelsMunch = QSpinBox(self)
        self._levelsMunch.setRange(0, 2147483647)
        self.layout().addRow(self._levelsMunchLabel, self._levelsMunch)

        self._paddingMunch = qt.QCheckBox("padding", self)
        self.layout().addWidget(self._paddingMunch)

        # vo parameters
        self._snrVOLabel = qt.QLabel("snr", self)
        self._snrVO = QDoubleSpinBox(self)
        self._snrVO.setMinimum(0.0)
        tooltip = "Ratio used to locate large stripes. Greater is less sensitive."
        self._snrVO.setToolTip(tooltip)
        self._snrVOLabel.setToolTip(tooltip)
        self.layout().addRow(self._snrVOLabel, self._snrVO)

        self._laSizeVOLabel = qt.QLabel("la_size", self)
        self._laSizeVO = QSpinBox(self)
        self._laSizeVO.setMinimum(0)
        tooltip = "Window size of the median filter to remove large stripes."
        self._laSizeVO.setToolTip(tooltip)
        self._laSizeVOLabel.setToolTip(tooltip)
        self.layout().addRow(self._laSizeVOLabel, self._laSizeVO)

        self._smSizeVOLabel = qt.QLabel("sm_size", self)
        self._smSizeVO = QSpinBox(self)
        self._smSizeVO.setMinimum(0)
        tooltip = "Window size of the median filter to remove small-to-medium stripes."
        self._laSizeVO.setToolTip(tooltip)
        self._smSizeVOLabel.setToolTip(tooltip)
        self.layout().addRow(self._smSizeVOLabel, self._smSizeVO)

        self._dimVOLabel = qt.QLabel("dimension", self)
        self._dimVO = QComboBox(self)
        self._dimVO.addItems(self._VO_DIMS)
        self.layout().addRow(self._dimVOLabel, self._dimVO)

        # sino mean deringer
        self._sigmaLowLabel = qt.QLabel("signal low", self)
        self._sigmaLow = QDoubleSpinBox(self)
        self._sigmaLow.setMinimum(0.0)
        self._sigmaHighLabel = qt.QLabel("signal high", self)
        self.layout().addRow(self._sigmaLowLabel, self._sigmaLow)

        self._sigmaHigh = QDoubleSpinBox(self)
        self._sigmaHigh.setMinimum(0.0)
        tooltip = (
            "sigma low and sigma high values are defining the standard deviation of "
            "gaussian(sigma_low) * (1 - gaussian(sigma_high)). \n"
            "High values of sigma mean stronger effect of associated filters."
        )
        self._sigmaHigh.setToolTip(tooltip)
        self._sigmaLow.setToolTip(tooltip)
        self.layout().addRow(self._sigmaHighLabel, self._sigmaHigh)
        # set up
        self.resetConfiguration()

    def resetConfiguration(self):
        self.setMethod(method=RingCorrectionMethod.MUNCH)
        self._levelsMunch.setValue(10)
        self._sigmaMunch.setValue(1.0)
        self._paddingMunch.setChecked(False)

        self._sigmaHigh.setValue(30.0)
        self._sigmaLow.setValue(0.0)

        self._snrVO.setValue(3.0)
        self._laSizeVO.setValue(51)
        self._smSizeVO.setValue(21)

    def getVoDim(self):
        if self._dimVO.currentText() == self._VO_DIMS[0]:
            return 1
        elif self._dimVO.currentText() == self._VO_DIMS[1]:
            return 2
        else:
            raise NotImplementedError

    def setVoDim(self, dim: int | str):
        if dim in ("1", 1, self._VO_DIMS[0]):
            self._dimVO.setCurrentText(self._VO_DIMS[0])
        elif dim in ("2", 2, self._VO_DIMS[1]):
            self._dimVO.setCurrentText(self._VO_DIMS[1])
        else:
            raise NotImplementedError(f"dim {dim} not handled")

    def getOptions(self) -> dict:
        if self.getMethod() is RingCorrectionMethod.NONE:
            return {}
        elif self.getMethod() is RingCorrectionMethod.MUNCH:
            return {
                "sigma": self._sigmaMunch.value(),
                "levels": self._levelsMunch.value(),
                "padding": self._paddingMunch.isChecked(),
            }
        elif self.getMethod() is RingCorrectionMethod.VO:
            return {
                "snr": self._snrVO.value(),
                "la_size": self._laSizeVO.value(),
                "sm_size": self._smSizeVO.value(),
                "dim": self.getVoDim(),
            }
        elif self.getMethod() in (
            RingCorrectionMethod.MEAN_DIVISION,
            RingCorrectionMethod.MEAN_SUBTRACTION,
        ):
            return {
                "filter_cutoff": (self._sigmaLow.value(), self._sigmaHigh.value()),
            }
        else:
            raise NotImplementedError

    def setOptions(self, options: dict) -> None:
        # handle munch propertoies
        if "sigma" in options:
            self._sigmaMunch.setValue(float(options["sigma"]))
        if "levels" in options:
            self._levelsMunch.setValue(int(options["levels"]))
        padding = options.get("padding")
        if padding is not None:
            self._paddingMunch.setChecked(padding in (True, 1, "1", "True"))
        # handle VO properties
        snr = options.get("snr")
        if snr is not None:
            self._snrVO.setValue(float(snr))
        la_size = options.get("la_size")
        if la_size is not None:
            self._laSizeVO.setValue(int(la_size))
        sm_size = options.get("sm_size")
        if sm_size is not None:
            self._smSizeVO.setValue(int(sm_size))
        dim = options.get("dim")
        if dim is not None:
            self.setVoDim(dim)
        # handle mean subtraction or division options
        filter_cutoff = options.get("filter_cutoff")
        if filter_cutoff is not None:
            low_pass, high_pass = filter_cutoff
            self._sigmaLow.setValue(float(low_pass))
            self._sigmaHigh.setValue(float(high_pass))

    def setMethod(self, method: RingCorrectionMethod):
        method = RingCorrectionMethod(method)
        self._method = method
        # handle munch options
        self._sigmaMunch.setVisible(method is RingCorrectionMethod.MUNCH)
        self._sigmaMunchLabel.setVisible(method is RingCorrectionMethod.MUNCH)
        self._levelsMunch.setVisible(method is RingCorrectionMethod.MUNCH)
        self._levelsMunchLabel.setVisible(method is RingCorrectionMethod.MUNCH)
        self._paddingMunch.setVisible(method is RingCorrectionMethod.MUNCH)
        # handle VO options
        self._snrVO.setVisible(method is RingCorrectionMethod.VO)
        self._snrVOLabel.setVisible(method is RingCorrectionMethod.VO)
        self._laSizeVO.setVisible(method is RingCorrectionMethod.VO)
        self._laSizeVOLabel.setVisible(method is RingCorrectionMethod.VO)
        self._smSizeVO.setVisible(method is RingCorrectionMethod.VO)
        self._smSizeVOLabel.setVisible(method is RingCorrectionMethod.VO)
        self._dimVO.setVisible(method is RingCorrectionMethod.VO)
        self._dimVOLabel.setVisible(method is RingCorrectionMethod.VO)
        # mean subtractions / division deringer
        self._sigmaLow.setVisible(
            method
            in (
                RingCorrectionMethod.MEAN_DIVISION,
                RingCorrectionMethod.MEAN_SUBTRACTION,
            )
        )
        self._sigmaLowLabel.setVisible(
            method
            in (
                RingCorrectionMethod.MEAN_DIVISION,
                RingCorrectionMethod.MEAN_SUBTRACTION,
            )
        )
        self._sigmaHigh.setVisible(
            method
            in (
                RingCorrectionMethod.MEAN_DIVISION,
                RingCorrectionMethod.MEAN_SUBTRACTION,
            )
        )
        self._sigmaHighLabel.setVisible(
            method
            in (
                RingCorrectionMethod.MEAN_DIVISION,
                RingCorrectionMethod.MEAN_SUBTRACTION,
            )
        )

    def getMethod(self) -> RingCorrectionMethod:
        return self._method


class TiltCorrection(qt.QGroupBox):
    """
    GroupBox dedicated to nabu TiltCorrection
    """

    sigChanged = qt.Signal()
    """Signal emit when parameters of the tilt options changed"""

    def __init__(self, text, parent=None, *args, **kwargs) -> None:
        super().__init__(text, parent, *args, **kwargs)
        self.setCheckable(True)
        self.setLayout(qt.QFormLayout())
        self._tiltManualRB = qt.QRadioButton("angle", self)
        self._angleValueSB = QDoubleSpinBox(self)
        self._angleValueSB.setRange(-360, 360)
        self._angleValueSB.setSuffix("°")
        self.layout().addRow(self._tiltManualRB, self._angleValueSB)

        self._autoManualRB = qt.QRadioButton("auto", self)
        self._autoModeCB = QComboBox(self)
        self._modes = {
            "1d-correlation": "auto-detect tilt with the 1D correlation method (fastest, but works best for small tilts)",
            "fft-polar": "auto-detect tilt with polar FFT method (slower, but works well on all ranges of tilts)",
        }
        for value, tooltip in self._modes.items():
            self._autoModeCB.addItem(value)
            idx = self._autoModeCB.findText(value)
            self._autoModeCB.setItemData(idx, tooltip, qt.Qt.ToolTipRole)
        self.layout().addRow(self._autoManualRB, self._autoModeCB)
        self._autoTiltOptions = qt.QLineEdit("", self)
        self._autoTiltOptions.setPlaceholderText("low_pass=1; high_pass=20 ; ...")
        self._autotiltOptsLabel = qt.QLabel("autotilt options")
        self._autotiltOptsLabel.setToolTip(
            """
        Options for methods computing automatically the detector tilt. \n
        The parameters are separated by commas and passed as 'name=value', for example: low_pass=1; high_pass=20. Mind the semicolon separator (;). \n
        For more details please see https://www.silx.org/pub/nabu/doc/apidoc/nabu.estimation.tilt.html#nabu.estimation.tilt.CameraTilt.compute_angle
        """
        )
        self.layout().addRow(self._autotiltOptsLabel, self._autoTiltOptions)
        ## `rotate_projections_center` option
        self._rotateProjectionCenter = RotateProjectionCenterWidget(self)
        self._rotateProjectionCenter.setToolTip(
            "\n".join(
                (
                    "Allows to correct detector tilt along beam orthogonal plan.",
                    "By default the center of rotation is the middle of each radio, i.e ((Nx-1)/2.0, (Ny-1)/2.0).",
                )
            )
        )
        self.layout().addRow(self._rotateProjectionCenter)

        # set up
        self._autoManualRB.setChecked(True)

        # connect signal / slot
        self._tiltManualRB.toggled.connect(self._updateVisibility)
        self._autoManualRB.toggled.connect(self._updateVisibility)

        self._tiltManualRB.toggled.connect(self._changed)
        self._autoManualRB.toggled.connect(self._changed)
        self._angleValueSB.valueChanged.connect(self._changed)

        self._updateVisibility()

    def _changed(self, *args, **kwargs):
        self.sigChanged.emit()

    def _updateVisibility(self):
        self._angleValueSB.setEnabled(self._tiltManualRB.isChecked())
        self._autoModeCB.setEnabled(self._autoManualRB.isChecked())
        self._autotiltOptsLabel.setVisible(self._autoManualRB.isChecked())
        self._autoTiltOptions.setVisible(self._autoManualRB.isChecked())

    def getTiltCorrection(self) -> tuple:
        """
        return (tilt value, autotilt options (if any))
        """
        if not self.isChecked():
            return "", ""
        elif self._tiltManualRB.isChecked():
            return self._angleValueSB.value(), ""
        else:
            return self._autoModeCB.currentText(), self._autoTiltOptions.text()

    def setTiltCorrection(
        self, tilt_correction: str, auto_tilt_options: str | None = None
    ) -> None:
        if tilt_correction in ("", None):
            self.setChecked(False)
        elif tilt_correction in self._modes.keys():
            self.setChecked(True)
            self._autoManualRB.setChecked(True)
            idx = self._autoModeCB.findText(tilt_correction)
            self._autoModeCB.setCurrentIndex(idx)
        else:
            self.setChecked(True)
            self._tiltManualRB.setChecked(True)
            self._angleValueSB.setValue(float(tilt_correction))
        if auto_tilt_options is not None:
            self._autoTiltOptions.setText(auto_tilt_options)

    def getRotateProjectionsCenter(self):
        if self.isChecked():
            return self._rotateProjectionCenter.getRotateProjectionsCenter()
        else:
            return None

    def setRotateProjectionsCenter(self, values):
        self._rotateProjectionCenter.setRotateProjectionsCenter(values=values)


class RotateProjectionCenterWidget(qt.QWidget):

    RANGE_MIN_VALUE = -sys.float_info.max
    RANGE_MAX_VALUE = sys.float_info.max

    def __init__(self, parent: qt.QWidget | None = None) -> None:
        super().__init__(parent)
        self.setLayout(qt.QHBoxLayout())
        self._label = qt.QLabel("rotate projections center")
        self.layout().addWidget(self._label)

        self._xQLE = QDoubleSpinBox(self)
        self._xQLE.setRange(
            RotateProjectionCenterWidget.RANGE_MIN_VALUE,
            RotateProjectionCenterWidget.RANGE_MAX_VALUE,
        )
        self._xQLE.setPrefix("x=")
        self._xQLE.setToolTip("Value should be in [-radio width, radio width]")
        self._xQLE.setSuffix("px")
        self.layout().addWidget(self._xQLE)

        self._yQLE = QDoubleSpinBox(self)
        self._yQLE.setRange(
            RotateProjectionCenterWidget.RANGE_MIN_VALUE,
            RotateProjectionCenterWidget.RANGE_MAX_VALUE,
        )
        self._yQLE.setPrefix("y=")
        self._xQLE.setToolTip("Value should be in [-radio heigh, radio heigh]")
        self._yQLE.setSuffix("px")
        self.layout().addWidget(self._yQLE)

    def getRotateProjectionsCenter(self) -> tuple | None:
        if self._xQLE.value() == 0 and self._yQLE.value() == 0:
            return None
        else:
            return (self._xQLE.value(), self._yQLE.value())

    def setRotateProjectionsCenter(self, values: tuple[float] | None):
        if values is None:
            self._xQLE.setValue(0.0)
            self._yQLE.setValue(0.0)
        elif not isinstance(values, (tuple, list)) and len(values) != 2:
            raise TypeError("values should be a tuple of two elements")
        else:
            self._xQLE.setValue(values[0])
            self._yQLE.setValue(values[1])
